structure AgentSmith :> COP =
struct

  structure T = Types
  structure S = Super
  structure G = Graph
  structure V = Vector
  structure A = Array
  structure M = Matrix

  exception Impossible

  type fact = T.nid * real
  type informant = T.bid

  exception ST of {foot_reach : M.connects, 
                   car_reach : M.connects,
                   slime : M.connects,
                   (* One set of beliefs for each world * cop. *)
                   historical_worlds : (informant * M.beliefs V.vector) list,
                   (* Historical data used to correct our predictions. *)
                   historical_facts : (informant * fact list A.array) list,
                   (* Used to project scores into the future. *)
                   future_worlds : M.beliefs V.vector,
                   (* Previous bank values. *)
                   bank_vals : (T.nid * int) list,
                   (* How much we trust each cop *)
                   trust : (T.bid * real) list,
                   (* Previous cop positions. *)
                   cop_positions : (T.pid * T.nid) list A.array,
                   (* Cops we plan to accuse. *)
                   accuse : T.pid list,
                   (* Bots we control. *)
                   controlling : T.pid list,
                   (* Temporary storage. *)
                   tmp : M.beliefs}

  val DEFAULT_LOOK_BEHIND = 10
  val LOOK_BEHIND = Params.param (Int.toString DEFAULT_LOOK_BEHIND)
                                (SOME ("-look_behind", "Smith look-behind")) 
                                "smith-look_behind"
  val DEFAULT_LOOK_AHEAD = 10
  val LOOK_AHEAD = Params.param (Int.toString DEFAULT_LOOK_AHEAD)
                                (SOME ("-look_ahead", "Smith look-ahead")) 
                                "smith-look_ahead"
  val DEFAULT_GAMMA = 70
  val GAMMA = Params.param (Int.toString DEFAULT_GAMMA)
                           (SOME ("-gamma", "Smith gamma weight (%)")) 
                           "smith-gamma"

  val FOOT_ONLY = Params.flag false (SOME ("-foot_only", 
                                           "Smith restricted to foot travel"))
                              "smith-foot_only"

  val DONT_ACCUSE = Params.flag false (SOME ("-dont_accuse", 
                                           "Smith doesn't accuse"))
                              "smith-dont_accuse"

  val DEFAULT_TRUST = 0.8

  val COP_POSITION_HISTORY = 14

  fun register () : S.register_msg * S.state = 
      (("AgentSmith", T.CopFoot), ST {foot_reach = M.dummy_connects,
                                      car_reach = M.dummy_connects,
                                      slime = M.dummy_connects,
                                      historical_worlds = nil,
                                      historical_facts = nil,
                                      future_worlds = V.fromList nil,
                                      bank_vals = nil,
                                      trust = nil,
                                      cop_positions = A.fromList nil,
                                      accuse = nil,
                                      controlling = nil,
                                      tmp = M.dummy_beliefs})

  fun lookup s ((s', x)::rst) = if s = s' then x else lookup s rst
    | lookup _ nil = raise Impossible

  fun world_skeleton (u, msg, s) : S.state = 
      let val g = S.get_graph u
          val len = G.length g
          val robberstart = G.robberstart g
          val bd = S.get_name u
          val look_behind = Params.asint DEFAULT_LOOK_BEHIND LOOK_BEHIND
          val look_ahead = Params.asint DEFAULT_LOOK_AHEAD LOOK_AHEAD

          val foot_reach = M.create_spread_connects g (G.Only T.CopFoot)
          val car_reach = M.create_spread_connects g (G.Only T.CopCar)
          val slime = M.create_spread_connects g (G.Only T.Robber)

          fun gray_beliefs length = A.tabulate (length, fn _ => 1.0 / 
                                                       (Real.fromInt length))
          val historical_worlds = List.map (fn bd => (bd, V.tabulate 
                                    (look_behind, fn _ => gray_beliefs len)))
                                           (S.get_cops_as_list u)

          val historical_facts = List.map (fn bd => (bd, A.tabulate 
                                            (look_behind, fn _ => nil)))
                                          (S.get_cops_as_list u)
          val future_worlds = V.tabulate (look_ahead, 
                                          (fn _ => M.new_beliefs len))
          val trust = List.map (fn bd => (bd, DEFAULT_TRUST))
                               (S.get_cops_as_list u)
          val cop_positions = A.tabulate (COP_POSITION_HISTORY,
                                          fn _ => nil)
      in
        ST {foot_reach = foot_reach,
            car_reach = car_reach,
            slime = slime,
            historical_worlds = historical_worlds,
            historical_facts = historical_facts,
            future_worlds = future_worlds,
            bank_vals = nil,
            trust = trust,
            cop_positions = cop_positions,
            accuse = nil,
            controlling = nil,
            tmp = M.new_beliefs len}
      end

  (* XXX Cut-and-paste from McGruff! *)
  fun inform (u, w_msg : S.world_msg, 
              ST {foot_reach, car_reach, slime, historical_worlds,
                  historical_facts, future_worlds, bank_vals,
                  trust, cop_positions, tmp, ...}) 
      : S.inform_msg * S.state = 
      let val (nd, pt) = S.get_my_loc u
          val smell = S.get_smell u
          val dist = case smell
                      of SOME (d) => d
                       | NONE => case pt of T.CopFoot => 2
                                          | T.CopCar => 1
                                          | _ => raise Impossible
          val reachable = G.reachable (S.get_graph u) G.Forward (G.Only pt)
                                      [nd] dist
          val nreachable = List.length reachable - 1 (* subtract self *)
          val robber = S.get_robber u
          val now = S.get_time u

          (* This is AgentSmith specific. *)
          val g = S.get_graph u
          val cur_time = ((S.get_time u) - 1) div 2
          val look_behind = Params.asint DEFAULT_LOOK_BEHIND LOOK_BEHIND
          val i = cur_time mod look_behind
          val me = S.get_name u
          val my_facts = lookup me historical_facts

          (* ROBBERIES *)
          val new_bank_vals = #bank_vals w_msg
          val sum_of_bank_vals = List.foldl (fn ((_, lt), tlt) => lt + tlt) 0
          val old_sum = sum_of_bank_vals bank_vals
          val new_sum = sum_of_bank_vals new_bank_vals

          val robberies = 
              if new_sum < old_sum then
                List.mapPartial (fn (nd, lt) =>
                                    List.find (fn (nd', lt') =>
                                                  nd = nd' andalso lt < lt')
                                              bank_vals)
                                new_bank_vals
                          else nil

          val robbery = case robberies of [(nd, _)] => SOME nd
                                        | nil => NONE
                     (* Robbery and bank transfer in the same turn means that
                      * we can't tell where the robbery occured. *)
                                        | _ => NONE 

          (* EVIDENCE *)
          val evidence = #evidence w_msg
          val evidence_informs = List.map (fn (nd, t) => (robber, nd, T.Robber,
                                                          t, 100))
                                          evidence

          (* Debugging *)
          val _ = if List.length evidence > 0 then Spew.spew "Found EVIDENCE:"
                  else ()
          val _ = List.app (fn (nd, t) =>
                               Spew.spew ((G.nid_to_name g nd) ^ " ("
                                          ^ (Int.toString nd) ^ ") "
                                          ^ (Int.toString t))) 
                           evidence
          (* *)


          (* Check to see if any cops have given themselves away by walking
           * over this evidence in the past. *)
          val accuse = List.map #1
                       (List.foldl 
                          (fn ((nd, t), outer_rst) => 
                              Util.ford ((t - 1) div 2) (cur_time - 1) nil
                              (fn (t', inner_rst) =>
                                  let val cops = 
                                          A.sub (cop_positions, 
                                                 t' mod COP_POSITION_HISTORY)
                                  in
                                    (List.filter (fn (pd, nd') => nd = nd') 
                                                 cops) @ inner_rst
                                  end)) 
                          nil
                          evidence)

          val accuse = if !DONT_ACCUSE then nil else accuse

          (* Debugging *)
          val _ = List.app (fn pd => Spew.spew ("I ACCUSE " ^ pd)) accuse
          (* *)

          val inform_msg =
              case robbery
               of SOME (nd) => [(robber, nd, T.Robber, now, 100)]
                | NONE => (case smell
                            of SOME (d) => List.map (fn nd =>
                                                   (robber, nd, T.Robber, now, 
                                                    100 div nreachable))
                                                    reachable
                             | NONE => List.map (fn nd => (robber, nd, 
                                                    T.Robber, now, ~100))
                                                (nd :: reachable))
          val inform_msg = inform_msg @ evidence_informs

          val controlling = List.map #2 (List.filter (fn (bd, _) => bd = me)
                                                     (#controlling w_msg))

          val s = ST {foot_reach = foot_reach, car_reach = car_reach,
                      slime = slime, historical_worlds = historical_worlds,
                      historical_facts = historical_facts,
                      future_worlds = future_worlds,
                      bank_vals = new_bank_vals, trust = trust,
                      cop_positions = cop_positions, accuse = accuse, 
                      controlling = controlling, tmp = tmp}
      in
        (inform_msg, s)
      end
    | inform _ = raise Impossible

  fun normalize_and_update a (ups : (int * real) list) =
      let val total = List.foldl (fn ((_, v), r) => v + r) 0.0 ups
          val rem = 1.0 - total
      in
        A.modifyi (fn (nd, v) => case List.find (fn (nd', _) => nd = nd') ups
                                   of SOME (_, v') => v'
                                    | NONE => rem * v)
                  a
      end

  fun pick_move u scores pd =
      let val g = S.get_graph u
          val (nd, pt) = S.get_player_loc (u, pd)

          val trans = if !FOOT_ONLY 
                      (* Note that bots we have taken control of may have
                       * been in a cop car -- we cannot force them to walk. *)
                      then case pt of T.CopFoot => G.Only T.CopFoot
                                    | _ => G.Any pt
                      else G.Any pt
          val steps = G.adjacent g G.Forward trans nd
          val step_scores = List.map (fn (nd, pt) => 
                                         ((nd, pt), A.sub (scores, nd))) steps

          (* Debugging 
          val _ = Spew.spew ("Picking from steps for " ^ (Int.toString pd))
          val _ = List.app (fn ((nd, _), sc) => Spew.spew
                                                    ((G.nid_to_name g nd) 
                                                     ^ " "
                                                     ^ (Real.toString sc)))
                           step_scores
          *)

          val ((nd, pt), _) = List.foldl (fn ((st, sc), (best_st, best_sc)) =>
                                             if sc > best_sc then (st, sc)
                                             else (best_st, best_sc))
                                         ((nd, pt), ~1000.0) step_scores
      in
        (nd, pt, pd)
      end

  (* XXX what about bogus informs? (e.g. where total certainty > 100) *)

  fun plan (u, fi_msg : S.from_inform_msg, 
            s as ST {foot_reach, car_reach, slime, historical_worlds, 
                     historical_facts, future_worlds, cop_positions, 
                     controlling, tmp, ...})
      : S.plan_msg * S.state =
      let val cur_time = ((S.get_time u) - 1) div 2
          val (nd, pt) = S.get_my_loc u
          val me = S.get_name u
          val robber = S.get_robber u
          val g = S.get_graph u

          val look_behind = Params.asint DEFAULT_LOOK_BEHIND LOOK_BEHIND

          (* First clear out space in the current slot. *)
          val _ = List.app (fn (_, facts) => 
                               A.update (facts, cur_time mod look_behind, nil))
                           historical_facts

          (* Because we know where cops are, we know where robbers are not. *)
          val my_facts = lookup me historical_facts
          val cop_pos_facts = List.map (fn (_, nd, _) => (nd, 0.0)) 
                                   (S.get_cop_locs u)
          val _ = A.update (my_facts, cur_time mod look_behind, 
                            cop_pos_facts @ (A.sub (my_facts, 
                                                    cur_time mod look_behind)))

          (* Save the informs too. *)
          val _ = List.app (fn (inf, informs) => 
                               let val facts = lookup inf historical_facts
                               in 

                                 List.app (fn (pd, nd, _, t, c) =>
                                   let val t = (t - 1) div 2 in
                                   if pd = robber andalso
                                      t > cur_time - look_behind andalso
                                      t <= cur_time then
                                     let val i = t mod look_behind
                                         val v = (Real.fromInt
                                                  (if c < 0 then 100+c else c))
                                                 / 100.0
                                         val rst = A.sub (facts, i) 
                                         (*
                                         val _ = Spew.spew ("Using inform " 
                                                         ^ (G.nid_to_name g nd)
                                                         ^ " ("
                                                         ^ (Int.toString nd) 
                                                         ^ ") "
                                                         ^ (Real.toString v))
                                          *)
                                     in
                                       A.update (facts, i, (nd, v)::rst)
                                     end
                                   else ()
                                   end) informs
                               end)
                           fi_msg


          (* Start with the oldest information about the robber, and propagate
           * information forward, using our facts to correct mistakes. *)
          fun propagate_facts a b facts =
              (M.multiply_with_dest slime b a;
               normalize_and_update b facts)

          fun print_facts ((nd, v)::fs) = Spew.spew ("Fact: " 
                                                     ^ (G.nid_to_name g nd)
                                                     ^ " ("
                                                     ^ (Int.toString nd) 
                                                     ^ ") "
                                                     ^ (Real.toString v))
            | print_facts nil = ()

          fun process_facts (inf, facts) =
             let val worlds = lookup inf historical_worlds in
               Util.for 0 (look_behind - 2)
               (fn i => let val j = (cur_time + i + 1) mod look_behind
                        in
                          (* Special case for the begining of the game. *)
                          if cur_time + i + 1 < look_behind andalso inf = me
                          then A.update (V.sub (worlds, 
                                                (j + 1) mod look_behind), 
                                         G.robberstart g, 1.0)
                          else propagate_facts (V.sub (worlds, j)) 
                                               (V.sub (worlds, (j + 1)
                                                             mod look_behind))
                                               (A.sub (facts, (j + 1)
                                                             mod look_behind))
                        end)
             end
          val _ = List.app process_facts historical_facts

          (* Debugging 
          val _ = Spew.spew ("History at time: " ^ (Int.toString cur_time))
          val _ = V.mapi (fn (i, world) => 
                             (Spew.spew (Int.toString i);
                              M.print_beliefs world)) 
                         (lookup me historical_worlds)
          *)

         (* In addition, we also save cop position explicitly so that we can
          * try to accuse turncoats. *)
          val _ = A.update (cop_positions, cur_time mod COP_POSITION_HISTORY,
                            (List.map (fn (pd, nd, pt) => (pd, nd)) 
                                      (S.get_cop_locs u)))

          (* Finally, aggregate information from different sources together. *)
          val _ = A.modify (fn _ => 0.0) tmp
          val _ = List.app (fn (bd, worlds) => 
                               let val world = V.sub (worlds, 
                                                  cur_time mod look_behind)
                                   val w = if bd = me then 2.0 else 1.0
                               in
                                 (* XXX Informant weighting? *)
                                 M.add_in_place_with_scale tmp world w
                               end)
                           historical_worlds

          (* Debugging
          val _ = Spew.spew ("Sum at time: " ^ (Int.toString cur_time))
          val _ = M.print_beliefs tmp
          *)

          (* Propogate this information into the future. *)
          val look_ahead = Params.asint DEFAULT_LOOK_AHEAD LOOK_AHEAD
          val _ = A.modifyi (fn (nd, _) => A.sub (tmp, nd))
                            (V.sub (future_worlds, 0))
          val _ = Util.for 0 (look_ahead - 2)
                  (fn t =>  M.multiply_with_dest 
                                slime
                                (V.sub (future_worlds, t + 1))
                                (V.sub (future_worlds, t)))

          (* Debugging 
          val _ = Spew.spew ("Future at time: " ^ (Int.toString cur_time))
          val _ = V.mapi (fn (i, world) => 
                             (Spew.spew (Int.toString i);
                              M.print_beliefs world))
                         future_worlds
          *)

          (* Now compute the transitive value of each node. *)
          val gamma = (Real.fromInt (Params.asint DEFAULT_GAMMA GAMMA)) / 100.0
          (* XXX This use of reach is an approximation. *)
          val reach = case pt of T.CopFoot => foot_reach 
                               | T.CopCar => car_reach
                               | T.Robber => raise Impossible
          val _ = Util.for 0 (look_ahead - 2)
                           (fn t => 
                               let val t' = look_ahead - t - 1
                                   val _ = M.multiply_with_dest
                                               reach
                                               tmp
                                               (V.sub (future_worlds, t'))
                               in
                                 A.modifyi (fn (nd, v) => 
                                               v + gamma * (A.sub (tmp, nd)))
                                           (V.sub (future_worlds, t' - 1))
                               end)

          (* Debugging
          val _ = Spew.spew ("Scores at time: " ^ (Int.toString cur_time))
          val _ = V.mapi (fn (i, world) => 
                             (Spew.spew (Int.toString i);
                              M.print_beliefs world))
                         future_worlds
          *)

          val scores = V.sub (future_worlds, 0)

          (* Debugging 
          val _ = Spew.spew ("Final at time: " ^ (Int.toString cur_time))
          val _ = M.print_beliefs scores
          *)

          val t = (S.get_time u) + 2
          val plan = if t <= 200 then
                       List.map ((fn (nd, pt, pd) => (pd, nd, pt, t))
                                 o (pick_move u scores))
                                (S.get_cops_as_list u)
                     else nil
      in 
        (plan, s)
      end
    | plan _ = raise Impossible

  (* narcissitic voting *)
  fun vote (u, fp_msg, s) : S.vote_msg * S.state =
      let val me = S.get_name u
          val others = List.filter (fn cop => cop<>me) 
                                   (List.map (fn (bd, _) => bd) fp_msg)
      in (me::others, s)
      end

  (* XXX Use vote result to guess future cop movements? *)
  fun move (u, _, s as ST {future_worlds, 
                           accuse, controlling, tmp, ...})
      : S.clean_cop_move_msg * S.state =
      let val cur_time = ((S.get_time u) - 1) div 2

          (* Take the best score of every reachable node. *)
          val scores = V.sub (future_worlds, 0)

          val _ = List.app (fn pd => Spew.spew 
                                         (pd ^ ", YOU ARE UNDER MY CONTROL")) 
                           controlling

          val moves = List.map (pick_move u scores) 
                               ((S.get_name u)::controlling)

      (*
      (* NON-MULTI-BOT MODE *)
          val moves = List.map (pick_move u scores) 
                               [(S.get_name u)]
       *)

      in
        ({moves = moves,
          offer = S.StraightArrow,
          accuse = accuse}, s)
      end
    | move _ = raise Impossible

  fun r_inform (u, _, s) : S.inform_msg * S.state = (nil, s)

  fun r_plan (u, _, s) : S.plan_msg * S.state = (nil, s)

  fun r_vote (u, fp_msg, s) : S.vote_msg * S.state =
      (List.map (fn (bd, _) => bd) fp_msg, s)

  fun r_results (u, _, s) : S.state = s

  fun r_move (u, _, s) : S.dirty_cop_move_msg * S.state
    = let val (nd, pt) = S.get_my_loc u
      in

        ([(nd, pt, S.get_name u)], s)
      end

  fun dirtify (u, s) = ({inform=inform, plan=plan, vote=vote, move=r_move,
                         r_inform=r_inform, r_plan=r_plan, r_vote=r_vote,
                         r_results = r_results}, s)

  fun new _ = {register=register, world_skeleton=world_skeleton, 
               dirtify=dirtify, inform=inform, plan=plan, vote=vote, move=move}

end
