(* Circler cop circles through a random tour of banks, chosen at startup.

   Also, he attempts to pursue if he smells a robber or finds evidence.

   He never informs anybody else of anything but his plans, and he ignores
   everybody else's plans. *)

structure Circler :> COP =
struct

  structure S = Super
  structure T = Types
  structure G = Graph

  exception Impossible

  exception State of { plan : G.nid option,
                       rotation : G.nid vector * int,
                       laststep : G.step
                     }

  val SEED = Params.param "0" (SOME ("-seed", "Circler seed")) 
                              "Circler-seed"

  (* XXX not sure if i'm doing this "correctly" -wjl *)
  val UNPDIST = Params.param "3" (SOME ("-unp-dist",
                                        "Circler unprotected distance")) 
                              "Circler-unp-dist"
  val unpdist = Params.asint 3 UNPDIST

  val CLUSTER_DIST = Params.param "0" (SOME ("-clust-dist",
                                             "Circler cluster distance"))
                                  "Circler-clust-dist"
  val clustdist = Params.asint 0 CLUSTER_DIST

  val debugplan = Params.flag false (SOME ("-debugplan",
                                           "Circler plan debug output"))
                                    "Circler-debugplan"

  fun unparse_step u (nid, ptype) =
    let val graph = S.get_graph u
    in
        ("(" ^ G.nid_to_name graph nid ^ ", "
        ^ Parsing.unparse_ptype ptype ^ ")")
    end

  fun spew u s = Spew.spew ("{" ^ S.get_name u ^ "@" 
                            ^ Int.toString (S.get_time u) ^ "} " ^ s)

  fun register () : S.register_msg * S.state =
        (("Circler", T.CopFoot), State {plan = NONE,
                                        rotation = (Vector.fromList [], 0),
                                        laststep = (0, T.CopFoot)})

  fun world_skeleton (u, (msg : S.world_skeleton_msg), State s) : S.state =
        let val seed = case Params.asint 0 SEED
                        of 0 => (case MLton.Random.seed () of
                                   NONE => 0w0
                                 | SOME x => x)
                         | x => Word.fromInt x
            val () = MLton.Random.srand seed
            val () = spew u ("random seed = " ^ Word.toString seed)
            val () = spew u ("unprotected distance = " ^ Int.toString unpdist)
            val () = spew u ("cluster distance = " ^ Int.toString clustdist)

            val graph = S.get_graph u
            val banknids = G.banks graph
            val banknids = Rand.permute_list banknids
            val rotation = Vector.fromList banknids
            val () = spew u ("bank tour =")
            val () = Vector.app (fn nid => spew u ("   " ^
                                            (G.nid_to_name graph nid)))
                                rotation
         in
             State
             {plan = #plan s,
              rotation = (rotation, 0),
              laststep = #laststep s}
         end
    
    | world_skeleton _ = raise Fail "bad state in circler:world_skeleton"

  fun inform (u, w_msg, s) = 
      let val (nd, pt) = S.get_my_loc u
          val smell = S.get_smell u
          val dist = case smell
                      of SOME (d) => d
                       | NONE => case pt of T.CopFoot => 2
                                          | T.CopCar => 1
                                          | _ => raise Impossible
          val trans = G.Any pt
          val reachable = G.reachable (S.get_graph u) G.Forward trans [nd] dist
          val nreachable = List.length reachable - 1 (* subtract self *)
          val robber = S.get_robber u
          val now = S.get_time u
      in
        case smell
         of SOME (d) => (List.map (fn nd => (robber, nd, T.Robber, now, 
                                             100 div nreachable))
                                  reachable, s)
          | NONE => (List.map (fn nd => (robber, nd, T.Robber, now, ~100))
                              (nd :: reachable), s)
      end

  (* PERF replace with nid set?  some sort of bool array? *)
  fun contains xs x = List.exists (fn y => y = x) xs
  fun union s1 s2 =
        List.foldl (fn (x, uni) => if contains uni x
                                    then uni
                                    else x::uni)
                    s1
                    s2
  fun inter s1 s2 =
        List.foldl (fn (x, int) => if contains s2 x
                                      andalso
                                      not (contains int x)
                                    then x::int
                                    else int)
                   []
                   s1
  fun diff s1 s2 =
        List.foldl (fn (x, dif) => if contains s2 x
                                    then dif
                                    else x::dif)
                    []
                    s1

  fun get_cop_nids u = map (fn (_, nid, _) => nid) (S.get_cop_locs u)

  fun cops_in_range u nid range =
    let val graph = S.get_graph u
        val localfootnids = G.reachable graph G.Backward
                                        (G.Any T.CopFoot) [nid] range
        val localcarnids = G.reachable graph G.Backward
                                       (G.Any T.CopCar) [nid] range
        val localnids = union localfootnids localcarnids
        val copnids = get_cop_nids u
        val localcopnids = 
            List.foldl (fn (cn, cns) =>
                            if contains localnids cn then (* cop is local *)
                                cn :: cns
                            else
                                cns)
                        []
                        copnids
        val ncops = List.length localcopnids
        val () = spew u (Int.toString ncops ^ " cops in range")
    in
        ncops
    end

  fun unprotected u nid = cops_in_range u nid unpdist = 0

  fun robbery_loc u =
    SOME (#1 (S.get_player_loc (u, (S.get_robber u))))
    handle S.Super _ => NONE

  fun spew_list u unparse str steps =
        if !debugplan then
            (
            spew u ("i am at " ^ unparse_step u (S.get_my_loc u));
            spew u ("and my " ^ str ^ " is this:");
            spew u "[--------------------- ";
            List.app (fn step => spew u (unparse step)) steps;
            spew u " ---------------------]\n"
            )
        else
            ()

  (* true if we're one of the two closest officers to loc *)
  fun close_to_loc u loc =
    let val cops = S.get_cop_locs u
        val (mynid, _) = S.get_my_loc u
        val graph = S.get_graph u
        val coppathlens =
            List.mapPartial (fn (pid, nid, ptype) =>
                               Option.map (fn sp => (nid, List.length sp))
                                (G.shortest_path graph (G.Any ptype) nid loc))
                            cops
        val coppathlens =
            ListUtil.sort (fn ((nid1, len1), (nid2, len2)) =>
                            Int.compare (len1, len2))
                          coppathlens
    in
        case coppathlens of
            (nid1, _) :: (nid2, _) :: _ => nid1 = mynid orelse nid2 = mynid
          | _ => false (* default to assuming we're not close if nobody can
                          reach the robbery location, which shouldn't happen *)
    end

  fun rotate tour idx : (G.nid * (G.nid vector * int)) =
    (Vector.sub (tour, idx),
     (tour, (idx + 1) mod Vector.length tour))

  fun robbery_decide u loc (plan, rotation) =
        let val graph = S.get_graph u
            val () = spew u ("robbery at " ^ G.nid_to_name graph loc ^ "!")
        in
            if close_to_loc u loc then
                (spew u "proceeding to the scene...";
                 (loc, rotation))
            else
                (spew u ("too far, pursuing alternate plan: "
                         ^ G.nid_to_name graph plan);
                 (plan, rotation))
        end

  (* invariant: after planning, #plan state <> NONE *)
  fun plan (u, msg, State state) : S.plan_msg * S.state =
      let val (plan, rotation) =
          case (robbery_loc u, state) of
            (NONE, {plan = NONE, rotation = (tour, idx), ...}) =>
                rotate tour idx
          | (NONE, {plan = SOME plan, rotation = (tour, idx), ...}) =>
                (plan, (tour, idx))
          | (SOME loc, {plan = SOME plan, rotation = (tour, idx), ...}) =>
                robbery_decide u loc (plan, (tour, idx))
          | (SOME loc, {plan = NONE, rotation = (tour, idx), ...}) =>
                robbery_decide u loc (rotate tour idx)
      in
        spew u ("made a plan: go to " ^ G.nid_to_name (S.get_graph u) plan);
        ([], State {laststep = #laststep state,
                    plan = SOME plan,
                    rotation = rotation})
      end

    | plan _ = raise Fail "bad state in Circler.sml:plan"

  fun vote (u, fp_msg, s) : S.vote_msg * S.state =
      (List.map (fn (bd, _) => bd) fp_msg, s)

  fun random_move u excluding =
      let val () = spew u "making a random move!"
          val mystep as (myloc, myptype) = S.get_my_loc u
          val mytrans = G.Any myptype
          val graph = S.get_graph u
          val reachable = G.adjacent graph G.Forward mytrans myloc
          val reachable = diff reachable excluding
          val () = spew_list u (unparse_step u) "adjacency list" reachable
          val len = List.length reachable
          val i = Rand.randN len
          val move = if len = 0 then mystep else List.nth (reachable, i)
          val () = spew u ("i am moving from " ^ unparse_step u mystep)
          val () = spew u ("              to " ^ unparse_step u move)
      in
          move
      end

  fun unparse_evidence u (nid, time) =
        G.nid_to_name (S.get_graph u) nid ^ " @ world " ^ Int.toString time

  fun random_evidence_move u excluding =
        let val () = spew u "found some evidence!  sleuthing about..."
            val evidence = S.get_evidence u
            val () = spew_list u (unparse_evidence u) "EVIDENCE" evidence
            val evidence = ListUtil.sort (fn ((nid1, t1), (nid2, t2)) =>
                                            Int.compare (t2, t1))
                                         evidence
            val (evnid, evtime) =
                    case evidence of
                        [] => raise Fail "Circler: no evidence"
                      | (n, t)::_ => (n, t)
            val graph = S.get_graph u
            val mystep as (myloc, myptype) = S.get_my_loc u
            val robber_turns = (S.get_time u - evtime) div 2
            val robber_nids = Graph.reachable graph G.Forward (G.Only T.Robber)
                                                    [evnid] robber_turns
            val adj_steps = Graph.adjacent graph G.Forward (G.Any myptype) myloc
            val adj_nids = List.map (fn (nid, _) => nid) adj_steps
            val adj_nids = diff adj_nids (List.map (fn (n,_) => n) excluding)
            val try_nids = inter robber_nids adj_nids
            val () = spew_list u (unparse_step u) "exclude list" excluding
            val () = spew_list u (G.nid_to_name graph) "nid possibilities" try_nids
            val len = List.length try_nids
            val i = Rand.randN len
            val try_nid = if len = 0 then NONE else SOME (List.nth (try_nids,i))
            val try_step = Option.map (ListUtil.Alist.get op= adj_steps) try_nid
            val try_step = Option.join try_step
            val () = Option.app (fn st => spew u ("trying step " ^ unparse_step u st)) try_step
        in
            case try_step of
                NONE => random_move u excluding
              | SOME s => s
        end
      
  fun clustered u =
    let val (myloc, myptype) = S.get_my_loc u
    in
        cops_in_range u myloc clustdist > 1
    end

  fun mymove (u, msg, s as State state) : (S.nid * S.ptype) * S.state =
    let val mystep as (myloc, myptype) = S.get_my_loc u
    in
        case state of
            {plan = SOME destnid, laststep, rotation} =>
                let val graph = S.get_graph u
                    val destpatho = Graph.shortest_path graph (G.Any myptype)
                                                        myloc destnid
                in
                    case destpatho of
                        (* couldn't reach our destination anymore --
                           shouldn't happen, but be safe *)
                        NONE => (mystep, State {plan = NONE,
                                                rotation = rotation,
                                                laststep = mystep})
                      | SOME [] => (* raise Fail "Circler invariant broken" *)
                        (* XXX apparently, this happens sometimes.. *)
                                    (mystep, State {plan = NONE,
                                                    laststep = mystep,
                                                    rotation = rotation})
                      | SOME [m] => (m, State {plan = NONE,
                                               laststep = mystep,
                                               rotation = rotation})
                      | SOME (m::_) =>
                            if S.get_evidence u <> [] then
                                (random_evidence_move u [laststep],
                                 State {plan = SOME destnid,
                                        rotation = rotation,
                                        laststep = mystep})
                            else if Option.isSome (S.get_smell u) then
                                let val () = spew u ("i smell crime!  avoiding "
                                                     ^ "my last step, "
                                                     ^ unparse_step u laststep)
                                in
                                    (random_move u [laststep],
                                     State {plan = SOME destnid,
                                            rotation = rotation,
                                            laststep = mystep})
                                end
                            else if clustered u then
                                (random_move u [],
                                 State {plan = SOME destnid,
                                        rotation = rotation,
                                        laststep = mystep})
                            else
                                (m,
                                 State {plan = SOME destnid,
                                        rotation = rotation,
                                        laststep = mystep})

                end
          | {plan = NONE, rotation, laststep} =>
                raise Fail "Circler: invariant broken: no plan"
      end

    | mymove _ = raise Fail "bad state in Circler.sml:mymove"

  (* hardcode first few moves for round 1 *)
  val hardcoded_steps : (string * T.ptype) vector =
    Vector.fromList [("55-and-kimbark", T.CopCar),
                     ("55-and-kenwood", T.CopCar),
                     ("54-and-kenwood", T.CopCar),
                     ("54-and-ridgewood", T.CopCar)]

  (* ... but execute them with a reusable function!@ *)
  fun mymove' (u, msg, s as State state) : (S.nid * S.ptype) * S.state =
        let val nth = S.get_time u div 2
            val graph = S.get_graph u
            fun to_step (str, ptype) = (G.name_to_nid graph str, ptype)
        in
            if nth < (Vector.length hardcoded_steps) then
                (to_step (Vector.sub (hardcoded_steps, nth)), s)
            else
                mymove (u, msg, s)
        end

     | mymove' _ = raise Fail "bad state in Circler.sml:mymove'"

  fun move (u, msg, s) : S.clean_cop_move_msg * S.state =
        let val ((nid, ptype), s') = mymove' (u, msg, s)
        in ({moves = [(nid, ptype, S.get_name u)],
             offer = Message.StraightArrow,
             accuse = []},
            s')
        end

  fun r_inform (u, _, s) : S.inform_msg * S.state = (nil, s)

  fun r_plan (u, _, s) : S.plan_msg * S.state = (nil, s)

  fun r_vote (u, fp_msg, s) : S.vote_msg * S.state =
      (List.map (fn (bd, _) => bd) fp_msg, s)

  fun r_move (u, _, s) : S.dirty_cop_move_msg * S.state
    = let val (nd, pt) = S.get_my_loc u
      in
        ([(nd, pt, S.get_name u)], s)
      end

  fun r_results (u, _, s) : S.state = s

  fun dirtify (u, s) = ({inform=inform, plan=plan, vote=vote, move=r_move,
                         r_inform=r_inform, r_plan=r_plan, r_vote=r_vote,
                         r_results=r_results}, s)

  fun new _ = {register=register, world_skeleton=world_skeleton,
               inform=inform, plan=plan, vote=vote, move=move,
               dirtify=dirtify}

end
