structure Francois :> ROBBER =
  struct

    structure T = Types
    structure S = Super
    structure G = Graph

    exception Impossible

    exception ST

    val RADIUS = Params.param "3" (SOME ("-radius", 
                                         "Francois scoring radius")) 
                              "francois-radius"

    fun contains xs x = (case List.find (fn x' => x = x') xs
                          of SOME _ => true
                           | NONE => false)

    fun register () = (("Francois", T.Robber), ST)

    fun inform (u, _, s) : S.inform_msg * S.state = (nil, s)

    fun plan (u, _, s) : S.plan_msg * S.state = (nil, s)

    fun vote (u, fp_msg, s) : S.vote_msg * S.state = 
        (List.map (fn (bd, _) => bd) fp_msg, s)

    fun bribe (u, _, s) : S.bribe_msg * S.state = (S.NoBribe, s)

    fun world_skeleton (u, ws_msg : S.world_skeleton_msg, s) = s

    fun move (u, _, s) =
        let val time = S.get_time u
            (* These weights should change depending on how many steps are
               considered. *)
            val COP_PENALTY = (100 + time) div 10
            val BANK_BONUS = 5

            (* Find the set of robber-reachable nodes. *)
            val g = S.get_graph u
            val (cur, pt) = S.get_my_loc u

            val _ = Spew.spew ("World " ^ (Int.toString (S.get_time u)))
            fun pl ((nd, v)::V) = (Spew.spew ("  " ^ (G.nid_to_name g nd) 
                                              ^ " = " ^ (Int.toString v));
                                   pl V)
              | pl nil = ()

            val nds = List.map #1 (G.adjacent g G.Forward 
                                              (G.Only T.Robber) cur)
            val V = List.map (fn nd => (nd, 0)) nds

            val _ = pl V

            (* Impose the cop-penalty on any node within radius of the given
               cop. *)
            fun consider_cop radius (pid, V) =
                let val (nd, pt) = S.get_player_loc (u, pid)
                    val reachable = G.reachable g G.Forward (G.Any pt) 
                                                [nd] radius
                    val contains = contains reachable
                in 
                  List.map (fn (nd, v) => if contains nd 
                                          then (nd, v - COP_PENALTY) 
                                          else (nd, v))
                           V
                end

            (* Give the bank-bonus to any node within "radius" of the given
               bank. *)
            fun consider_bank radius ((nd, _, _, _), V) =
                let val reachable = G.reachable g G.Backward (G.Only T.Robber)
                                                [nd] radius
                    val contains = contains reachable
                in 
                  List.map (fn (nd, v) => if contains nd
                                          then (nd, v + BANK_BONUS) 
                                          else (nd, v))
                           V
                end

            (* Get lists of cops and banks. *)
            val cops = (S.get_cops_as_list u)
            (* XXX Banks are fixed for this rounds, so consider precomputing
               scores for banks. *)
            val banks = (S.get_banks_as_list u)

            fun repeat 0 V = V
              | repeat radius V =
                (* Update scores for nodes "radius" steps from cops/banks. *)
                let val V = List.foldl (consider_cop radius) V cops
                    val V = List.foldl (consider_bank radius) V banks
                in
                  pl V;
                  repeat (radius - 1) V
                end

            (* Consider nodes up to RADIUS steps away (and reconsider nodes 
               N steps away for every value of radius >= N). *)
            val V = repeat (Params.asint 1 RADIUS) V

            val _ = pl V

            (* Find the best score of all adjacent nodes. *)
            fun find_max ((nd, v), (best_nd, best_v)) =
                if v > best_v then (nd, v) else (best_nd, best_v)

            val (nxt, v) = List.foldl find_max (cur, ~100) V
        in
          ({move = nxt, result = S.BribeNone}, s)
        end

    fun new _ : Super.robber =
        { register = register,
          world_skeleton = world_skeleton,
          inform = inform, plan = plan, vote = vote, bribe = bribe,
          move = move }
  end
