structure Neo :> ROBBER = 
struct  

  structure T = Types
  structure S = Super
  structure G = Graph
  structure M = Matrix
  structure V = Vector
  structure A = Array

  exception Impossible

  exception ST of {(* Propagates information based on robber-reachability. *)
                   reach : M.connects, 
                   (* Propagates information based on cops-starting-on-foot. *)
                   foot_heat : M.connects, 
                   (* Propagates information based on cops-starting-in-car. *)
                   car_heat : M.connects,
                   (* Temporary storage. *)
                   tmp : M.beliefs,
                   (* Penalties for each node based on low degree. *)
                   degree : M.beliefs,
                   (* Beliefs about cops-starting-on-foot.  One entry for the
                    * current world and one for each step into the future. *)
                   foot_worlds : M.beliefs V.vector,
                   (* Beliefs about cops-starting-in-car. Indexes as above. *)
                   car_worlds : M.beliefs V.vector}

  val DEFAULT_LOOK_AHEAD = 12
  val LOOK_AHEAD = Params.param (Int.toString DEFAULT_LOOK_AHEAD)
                                (SOME ("-look_ahead", "Neo look-ahead")) 
                                "neo-look_ahead"
  val DEFAULT_GAMMA = 30
  val GAMMA = Params.param (Int.toString DEFAULT_GAMMA)
                           (SOME ("-gamma", "Neo gamma weight (%)")) 
                               "neo-gamma"
  val DEFAULT_BANK_VALUE = 10
  val BANK_VALUE = Params.param (Int.toString DEFAULT_BANK_VALUE)
                                (SOME ("-bank_value", "Neo bank value (%)")) 
                                "neo-bank_value"

  fun register () : S.register_msg * S.state =
      (("Neo", T.Robber), ST {reach = M.dummy_connects, 
                              foot_heat = M.dummy_connects,
                              car_heat = M.dummy_connects,
                              tmp = M.dummy_beliefs,
                              degree = M.dummy_beliefs,
                              foot_worlds = V.fromList [M.dummy_beliefs],
                              car_worlds = V.fromList [M.dummy_beliefs]})

  fun world_skeleton (u, _, _) : S.state = 
      let val g = S.get_graph u
          val len = G.length g
          val look_ahead = Params.asint DEFAULT_LOOK_AHEAD LOOK_AHEAD
      in
        (* Reachability uses robber transport since it is used to determine
         * where we can go. *)
        ST {reach = M.create_reach_connects g (G.Only T.Robber),
            (* Heat uses cop transport since it is used to determine where
             * the cops might be. *)
            foot_heat = M.create_spread_connects g (G.Any T.CopFoot),
            car_heat = M.create_spread_connects g (G.Any T.CopCar),
            (* Build empty arrays for each of the belief sets. *)
            tmp = M.new_beliefs len,
            degree = A.tabulate (len, 
                   fn nd => (Real.fromInt (List.length (G.adjacent g 
                                    G.Forward (G.Only T.Robber) nd)))
                                 / 6.0),
            foot_worlds = V.tabulate (look_ahead, fn _ => M.new_beliefs len),
            car_worlds = V.tabulate (look_ahead, fn _ => M.new_beliefs len)}
      end

  fun inform (u, _, s) : S.inform_msg * S.state = (nil, s)

  fun plan (u, _, s) : S.plan_msg * S.state = (nil, s)

  fun vote (u, fp_msg, s) : S.vote_msg * S.state = 
      (List.map (fn (bd, _) => bd) fp_msg, s)

  fun bribe (u, _, s) : S.bribe_msg * S.state = (S.NoBribe, s)

  fun move (u, _, s as ST {reach, foot_heat, car_heat, tmp, degree,
                           foot_worlds, car_worlds}) 
      : S.robber_move_msg * S.state
    = let val g = S.get_graph u
          val cops = S.get_cop_locs u
          val banks = S.get_banks_as_list u
          val cur_foot_world = V.sub (foot_worlds, 0)
          val cur_car_world = V.sub (car_worlds, 0)

          (* Clear the cop probabilities for the current world. *)
          val _ = M.clear_beliefs cur_foot_world
          val _ = M.clear_beliefs cur_car_world
          (* Set the probabilities for the current world based on the known
           * locations of cops. A value of ~n means we expect to see n cops. *)
          val _ = List.app (fn (_, nd, pt) => 
                               let val cur_world = 
                                       case pt of T.CopFoot => cur_foot_world
                                                | T.CopCar => cur_car_world
                                                | T.Robber => raise Impossible
                               in
                                 A.update (cur_world, nd, 
                                           A.sub (cur_world, nd) - 1.0)
                               end)
                           cops

          (* Propagate this information into the FUTURE. *)
          val look_ahead = Params.asint DEFAULT_LOOK_AHEAD LOOK_AHEAD
          fun propagate worlds heat =
              V.appi (fn (t, world) => if t = 0 then ()
                                       else M.multiply_with_dest heat
                                                                 world
                                                       (V.sub (worlds, t - 1)))
                     worlds
          val _ = propagate foot_worlds foot_heat
          val _ = propagate car_worlds car_heat

          (* Debugging *)
          val _ = Spew.spew "Foot-future:"
          val _ = M.print_beliefs (V.sub (foot_worlds, look_ahead - 1))
          val _ = Spew.spew "Car-future:"
          val _ = M.print_beliefs (V.sub (car_worlds, look_ahead - 1))
          (* *)

          (* Now add information about the banks.  Banks don't move, so we
           * just add a constant ("add" because banks are good) to each node
           * in each world where a bank appears.  We only add the bank bonus 
           * to one set of beliefs since we will combine the sets of beliefs in
           * the next step. *)
          val bank_value = (Real.fromInt (Params.asint DEFAULT_BANK_VALUE
                                                       BANK_VALUE)) / 100.0
          val _ = List.app (fn (nd, _, _, _) => 
                           Util.for 0 (look_ahead - 1)
                           (fn t => let val cur_world = V.sub (foot_worlds, t)
                                    in
                                      A.update (cur_world, nd, 
                                            A.sub (cur_world, nd) + bank_value)
                                    end))
                           banks
          (* Finally, aggregate information about foot-cops and car-cops
           * into a single set of beliefs.  We choose to aggregate into the
           * foot worlds, but there is no particular reason for this. *)
          val worlds = foot_worlds
          val cur_world = V.sub (worlds, 0)
          val _ = V.appi (fn (t, world) => 
                             (M.add_in_place world (V.sub (car_worlds, t));
                              M.add_in_place world degree))
                         worlds

          (* Two cops is no worse than one! *)
          (* fun ceil' a = A.modify (fn v => if v < ~1.0 then ~1.0 else v) a
          val _ = V.app ceil' worlds *)

          (* Debugging *)
          val _ = Spew.spew "Initial scores:"
          val _ = V.app (fn world => M.print_beliefs world) worlds
          (* *)

          (* Compute the score-over-time for each node in the graph. We do 
           * this in-place, starting from the most distant world and working
           * back to the present. *)
          val gamma = (Real.fromInt (Params.asint DEFAULT_GAMMA GAMMA)) / 100.0
          val _ = Util.for 0 (look_ahead - 2)
                  (fn t => let val t' = look_ahead - t - 1
                               val _ = M.multiply_with_dest reach
                                                            tmp
                                                         (V.sub (worlds, t'))
                           in
                             A.modifyi (fn (nd, v) => 
                                           v + gamma * (A.sub (tmp, nd)))
                                       (V.sub (worlds, t' - 1))
                           end)
          (* Debugging *)
          val _ = Spew.spew "Reachability scores:"
          val _ = V.app (fn world => M.print_beliefs world) worlds
          (* *)

          (* Take the best score of every reachable node. *)
          val (nd, pt) = S.get_my_loc u
          (* val nodes = G.reachable g G.Forward (G.Only pt) [nd] 1 *)
          val nodes = List.map #1 (G.adjacent g G.Forward (G.Only pt) nd)
          val node_scores = List.map (fn nd => (nd, A.sub (cur_world, nd)))
                                     nodes

          (* Debugging *)
          val _ = Spew.spew ("Time: " ^ (Int.toString (S.get_time u)))
          val _ = List.app (fn (nd, sc) => Spew.spew ((G.nid_to_name g nd) 
                                                      ^ " "
                                                      ^ (Real.toString sc)))
                           node_scores
          (* *)

          val (nd, _) = List.foldl (fn ((nd, sc), (best_nd, best_sc)) =>
                                       if sc > best_sc then (nd, sc)
                                       else (best_nd, best_sc))
                        (nd, ~1000.0) node_scores
      in
        ({move = nd, result = S.BribeNone}, s)
      end
    | move _ = raise Impossible


  fun new _ = {register=register, world_skeleton=world_skeleton, 
               inform=inform, plan=plan, vote=vote, bribe=bribe, move=move}
end
