(* I am Optimus Prime!
   I am Optimus Prime!
   I am Optimus Prime!
   I am Optimus Prime! *)

structure OptimusPrime :> ROBBER =
  struct

    structure T = Types
    structure S = Super
    structure G = Graph

    open Transform

    exception OptimusPrime of string

    val HEATFACTOR = 4
        
    (* number of robber moves to search deep
       (minus one)
       
       This is a degenerate iterative deepening
       strategy:

       We first search to SAFE_DEPTH, which we
       assume will finish in the allotted time.
       Then we search to OPTIMISTIC_DEPTH, which
       we hope will finish in the allotted time. *)
    val OPTIMISTIC_DEPTH = 4 (* XXX change to 5 *)
    val SAFE_DEPTH = 2

    (* number of seconds to search the game tree before returning an answer.
       should be less than 5! *)
    val TIMELIMIT = Time.fromReal 4.0

    val ALTERNATE_EVAL = Params.flag false (SOME ("-alternate-eval",
                                                  "Optimus Prime uses alternate evaluation"))
                                     "optimus-alternate-eval"

    val checkpoint = ref (fn () => ())
    val best_move : int option ref = ref NONE

    exception ST of
       { graph : Graph.graph,
         transfers : (Graph.nid * int) list,
         evidence : (Graph.nid * int) list,
         oldheat : RobberHeatTransform.cops_asheat,
         lastworld : (RobberHeatTransform.cops_asheat, exactrobber) gamestate option }

    exception NullPointer (* ;) *)
    (* start with bad exn, since we can't initialize yet *)
    fun register () = (("OptimusPrime", T.Robber), NullPointer)

    fun world_skeleton (u, ws_msg : S.world_skeleton_msg, _) = 

        let val g = #graph ws_msg
            val cold = Vector.tabulate(Graph.length g,
                                       fn _ => NONE)
        in
            ST { graph = g, transfers = nil, evidence = nil,
                 oldheat = { foot = cold,
                             car = cold }, lastworld = NONE }
        end
        

    (* evalrec d st move k
       
       evaluate a move for a gamestate (on robber's turn) recursively. 
       larger results are better. Call the continuation k on the
       result. *)
    fun evalrec cxp d (st as GS{time=198, loot, temp, ...}) move k =
        let 
            val stres = TransformByRobber.transform st move
            val stres = RobberHeatTransform.transform stres
            val opts = TransformByRobber.enumerate stres
        in
            (* end state. all we care about here is loot and temp. *)
            k (real (loot - 10 * temp))
        end
      | evalrec _ 0 (st as GS{time,...}) move k =
        let
            (* val stres = TransformByRobber.transform st move *)
            val evaluate = if !ALTERNATE_EVAL then EvaluateRobber'.evaluate
                                              else EvaluateRobber.evaluate
        in
            (* k (evaluate stres) (* XXX stres or st?  which works better? *) *)
            k (evaluate st)
        end
      | evalrec cxp d st move k =
        let
            val stres = TransformByRobber.transform st move
            val stres = RobberHeatTransform.transform stres
            val opts = TransformByRobber.enumerate stres

            fun try nil kl = kl nil
              | try (m::rest) kl =
                evalrec cxp (d - 1) stres m
                (fn res =>
                 let
                     fun go () =
                         try rest (fn l => kl (res::l))
                 in
                     (* PERF maybe don't checkpoint if depth
                        is too low? *)
                     cxp := go;
                     go ()
                 end)

            (* we'll get to choose, so just use best score
               (maximax) *)
            fun donelist l =
                k (ListUtil.max Real.compare l)

        in
            try opts donelist
        end

    (* val evalrec = fn x => (evalrec x handle e => raise e) *)

    fun computebest cxp depth start k =
        let 
                
            val opts = TransformByRobber.enumerate start

            (* search *)
            fun try nil kl = kl nil
              | try (m::rest) kl =
                evalrec cxp depth start m
                  (fn res =>
                   let 
                       fun go () =
                           try rest (fn l => kl ((res, m)::l))
                   in
                       cxp := go;
                       go ()
                   end)

            fun donelist l =
                let
                    val (_, bestmove) = 
                        (* always at least one, because of self loop *)
                        ListUtil.max (ListUtil.byfirst Real.compare) l
                in
                    k bestmove
                end
        in
            try opts donelist
        end

    (* find the best move, first by quickly searching to depth SAFE_DEPTH,
       then by trying to get to depth OPTIMISTIC_DEPTH *)
    fun twophase resref cxp start =
      let in
       Spew.spew "Opti running twophase...";
        computebest cxp SAFE_DEPTH start
        (fn res =>
         let in
             resref := SOME res;
             Spew.spew "Opti: setting resref for the first time";
             computebest cxp OPTIMISTIC_DEPTH start
               (fn res =>
                let in
                    Spew.spew "Opti: setting resref for the second time";
                    resref := SOME res;
                    cxp := (fn () => ())
                end)
         end)
      end

    fun move (u, offered_cops_msgs_unused, 
              ST { graph, 
                   transfers,
                   evidence,
                   oldheat,
                   lastworld = SOME start }) =
        let val _ = Engine.timeLimit (TIMELIMIT, !checkpoint)
            val chosen_move =
                case !best_move of
                    NONE =>
                        let val (my_nid, _) = Super.get_my_loc u
                        in
                            Spew.spew "Opti choosing dumb move!  bad!";
                            my_nid
                        end
                  | SOME nid => nid

            val newstate = TransformByRobber.transform start chosen_move
            val newstate = RobberHeatTransform.transform newstate

            val GS{transfers, evidence, cops, ...} = newstate
        in
            ({ move = chosen_move, result = S.BribeNone },
             ST { graph = graph,
                  transfers = transfers,
                  evidence = evidence,
                  oldheat = cops,
                  lastworld = NONE })
        end
      | move _ = raise OptimusPrime "driver error: wrong state (no lastworld?)"

    (*
    val move = fn x => (move x handle e => raise e)
    *)

    (* XXX no-ops! replace these with the good stuff. *)
    (* but here we have to remember the world message for later.. *)

    (* IMO inform is totally useless, since both the robbers and the cops
       know exactly where everyone is. We don't need to or want to
       trick anyone (maybe at the end of the round we might want our
       helper bots to get caught so that they don't share our loot?),
       so what is the point? *)
    fun inform (u, wm as {bank_vals, time, ...},
                s as ST {graph, transfers, evidence, oldheat, ... }) 
        : S.inform_msg * S.state = 
        let val (me, _) = S.get_my_loc u

            val () = Spew.spew ("Opti: time = " ^ Int.toString (#time wm))

            (* null hypothesis: robber could be anywhere *)
            val flatslime = 
                Vector.tabulate(Graph.length graph, 
                                fn _ => RobberHeatTransform.SLIMEFACTOR)

            val cops = 
                List.mapPartial (fn (_, n, T.CopFoot) => SOME (n, false)
                                 |  (_, n, T.CopCar) => SOME (n, true)
                                 | _ => NONE) (S.get_player_locs u)

            val num = Graph.length graph

            fun oldslime car i =
                let val a = if car 
                            then #car oldheat
                            else #foot oldheat
                in
                    case Vector.sub(a, i) of
                        NONE => flatslime
                      | SOME (_, v) => v
                end

            val heat =
                {foot = 
                 Vector.tabulate
                 (num,
                  fn i =>
                  let val ncops =
                      length (List.filter (fn (x, false) =>
                                           x = i | _ => false) cops)
                      * num * HEATFACTOR
                  in
                      if ncops > 0
                      then SOME (ncops, oldslime false i)
                      else NONE
                  end),

                 car = 
                 Vector.tabulate
                 (num,
                  fn i =>
                  let val ncops =
                      length (List.filter (fn (x, true) => x = i
                                           | _ => false) cops)
                      * num * HEATFACTOR
                  in
                      if ncops > 0
                      then SOME (ncops, oldslime true i)
                      else NONE
                  end) }

            (* initialize gamestate. *)
            val start 
                : (RobberHeatTransform.cops_asheat, 
                   Transform.exactrobber) Transform.gamestate
                =
                GS { skel = graph,
                     time = S.get_time u,
                     loot = S.get_loot u,
                     bank_vals = bank_vals,
                     temp = 0, (* safe for now *)
                     cops = heat,
                     robber = me,
                     (* we have to remember these ourselves,
                        since the world does not show them *)
                     evidence = evidence,
                     transfers = transfers }

            val () =
                case time of
                    0 => best_move := SOME (Graph.name_to_nid graph "54-and-dorchester")
                  | _ =>
                    let val () = best_move := NONE
                        fun go () = twophase best_move checkpoint start
                        val () = checkpoint := go
                        val _ = Engine.timeLimit (TIMELIMIT, go)
                    in
                        ()
                    end
        in
            (nil, 
             ST { graph = graph,
                  transfers = transfers, 
                  evidence = evidence, 
                  oldheat = oldheat,
                  lastworld = SOME start })
        end

      | inform (_, _, _) = raise OptimusPrime "wrong state for inform"

    (* no purpose to plans, so we ignore them *)
    fun plan (u, _, s) : S.plan_msg * S.state =
        let val _ = Engine.timeLimit (TIMELIMIT, !checkpoint)
        in
            (nil, s)
        end

    fun vote (u, fp_msg, s) : S.vote_msg * S.state = 
        let val _ = Engine.timeLimit (TIMELIMIT, !checkpoint)
        in
            (List.map (fn (bid, _) => bid) fp_msg, s)
        end
        
    fun bribe (u, vres_msg, s) : S.bribe_msg * S.state =
        let val _ = Engine.timeLimit (TIMELIMIT, !checkpoint)
        in
            (* XXX need to save a bribe message, too -- if we want to push,
               we have to say S.Bribe here! *)
            (S.NoBribe, s)
        end

    fun new _ : Super.robber =
        { register = register,
          world_skeleton = world_skeleton,
          move = move,
          inform = inform,
          plan = plan,
          vote = vote,
          bribe = bribe }
  end
