
(* This is an "inverse solver", that is, a program that generates puzzles.

   It works by expanding the entire state graph and then finding the
   most distant ("hardest") state that satisfies a certain predicate
   (such as being a "solved" state). 

   Given a correct state generator, this basic framework can work for
   many kinds of puzzles. Currently it is set up for sliding "gold"
   blocks puzzles where all blocks are indistinguishable and end up on
   certain target squares, and sokoban-like levels with a fixed number
   of grey blocks. Both of these work by a reverse method, starting
   from the solved state and working backwards to find the most distant
   state from which the solved state can be reached.

   There is also an experimental "Escape" solver that implements more
   tiles but is much much less efficient. This is a "forward" solver,
   which works by taking a starting configuration and finding a (most
   distant) solved state (player on exit). Here, the length of this
   solution is not indicative of the difficulty of the level, so it
   is mainly useful for finding solutions to levels, rather than
   generating difficult ones. *)

(* The inverse solver is parametrized on the "game", which defines the
   state space and its transition functions. A game has the following
   signature: *)

signature GAME =
sig
    type state

    val hash : state -> Word32.word
    val eq : state * state -> bool

    (* app_successors (f, s)
       apply f to every successor to s, along with its distance
       from the source *)
    val app_successors : (((state * int) -> unit) * state) -> unit
        
    (* the inital state *)
    val initial : state

    val tostring : state -> string

    (* short code *)
    val tocode : state -> string

    (* we'll find only paths to accepting states, according to
       this notion of accepting *)
    val accepting : state -> bool

    (* if true, then return the shortest path(s). if false, return 
       the longest. *)
    val findshortest : bool

    (* should the state list be reversed when printing? *)
    val reverse : bool
end


functor Search(G : GAME) =
struct

    structure H = HeapFn(type priority = int
                         val compare = Int.compare)

    val NBEST = 3

    exception Search of string

    (* make hash table 
       PERF: better if we can guess the approximate table size... *)

    type hentry = { (* XXX could save mem at expense of time by getting
                       this from the handle instead *)
                    me : G.state,

                    (* best predecessor so far (closest to initial state).
                       if equal to me, then this is the initial state. *)
                    pred : G.state,

                    (* distance between me and the predecessor state *)
                    (* can be optimized out in the case that it is constant, eg, 1 *)
                    dist : int,

                    (* total depth *)
                    (* can be optimized out when we search the path to compute depth *)
                    depth : int,
                    
                    (* handle into priority queue; might have gone bad
                       (test with H.valid) *)
                    hand : H.hand }

    val HASH_SIZE = 800011

    val table = Array.array (HASH_SIZE, nil : hentry list)
    val h = H.empty () : G.state H.heap

    fun indexof s =
        let
            val h = G.hash s
            val i = Word32.mod (h, Word32.fromInt HASH_SIZE)
        in
            Word32.toIntX i
        end

    (* ignores depth field. if this is called instead of just reading
       depth, then the depth field will never be used, so it should be
       optimized out, saving some memory. Note: this function is very
       expensive! If used, it often accounts for a majority of the
       execution time, when we search for the best state at the end. *)
    fun depth_path acc (s, pred) =
        if G.eq (s, pred)
        then acc (* initial *)
        else
            let
                val i = indexof pred
                val l = Array.sub(table, i)
                fun go nil = raise Search "path error!"
                  | go ({me=me', pred=pred', dist, ...}::rest) =
                    if G.eq (me', pred)
                    then (* found it *)
                         depth_path (acc + dist) (me', pred')
                    else go rest
            in
                go l
            end

    local
        val scount = ref 0
    in
        fun seen s =
            let in
                scount := !scount + 1;
                if (!scount mod 50000 = 0)
                then 
                    let in
                        TextIO.output (TextIO.stdErr, 
                                       "--------------------------------------------------\n");
                        TextIO.output (TextIO.stdErr, "States seen: " ^ Int.toString (!scount) ^ 
                                       "  Frontier size: " ^ Int.toString (H.size h) ^
                                       "\n");
                        (* include depth? *)
                        TextIO.output (TextIO.stdErr, G.tostring s ^ "\n")
                    end
                else ()
            end
        fun nseen () = !scount 
    end

    (* visit s', coming from s and with depth d *)
    fun visit d dist s s' =
        let val i = indexof s'
            val l = Array.sub(table, i)
        in
            (* have we seen it at all yet? *)
            case List.find (fn {me, ...} => G.eq (me, s')) l of
                (* no? then put it in the frontier *)
                NONE => 
                    let val hand = H.insert h d s'
                    in
                        (* and in the hash table... *)
                        Array.update(table, i, {me=s', pred=s, hand=hand, dist=dist, depth=d} :: l);
                        seen s
                    end
              | SOME {me, pred, hand, dist=_, depth=oldd} => 
                    if H.valid hand
                    then (* still in frontier, but maybe we need to
                            update it. *)
                        let (* val (oldd, _) = H.get h hand *)
                        in
                            if d < oldd
                            then 
                                let in
                                    H.adjust h hand d;
                                    Array.update(table, i, {me=s', pred=s, hand=hand, dist=dist, 
                                                            depth=d} ::
                                                 List.filter (fn {me, ...} => not (G.eq(me, s'))) l)
                                end
                            else ()
                        end
                    else () 
                        (* PERF this check is just for sanity *)
                      (*
                      let val best = depth 0 (me, pred)
                      in
                      if d < best
                      then raise Search "dijkstra was wrong!"
                      else ()
                      end
                      *)

        end


    fun loop () =
        case H.min h of
            NONE => () (* done -- frontier empty *)
          | SOME (d, s) =>
                let in
                    G.app_successors
                    (fn (s', dist) => visit (d + dist) dist s s', s);
                    loop ()
                end

    val () = visit 0 0 G.initial G.initial

    val () = loop ()

    val () = print ("Total states: " ^ Int.toString (nseen ()) ^ "\n")
    val () = TextIO.output (TextIO.stdErr, "Total states: " ^ Int.toString (nseen ()) ^ "\n")

    val naccept = ref 0

    val furthest =
        let
            val () = print "done! finding furthest states...\n";

            val best = ref (nil : ((G.state * G.state) * int) list)

            fun lookat {me, pred, depth=d, ...} =
              if G.accepting me
              then
                let
                  val cmp = if G.findshortest 
                          then ListUtil.bysecond Int.compare
                          else ListUtil.Sorted.reverse (ListUtil.bysecond Int.compare)

                  val l =
                    ListUtil.Sorted.insert cmp (!best) ((me, pred), d)
                in
                  naccept := !naccept + 1;
                  (* but only keep the five best *)
                  if length l > NBEST
                  then best := List.take (l,NBEST)
                  else best := l
                end
              else ()
        in
            Array.app (app lookat) table;
            !best
        end

    val () = print ("Total accepting: " ^ Int.toString (!naccept) ^ "\n")
    val () = TextIO.output (TextIO.stdErr, "Total accepting: " ^ 
                            Int.toString (!naccept) ^ "\n")

    (* and print them. *)
    val () = 
        let

            fun indent n s =
                StringUtil.replace "\n" ("\n" ^ StringUtil.tabulate n #" ") s

            fun solution acc (s, pred) =
                if G.eq (s, pred)
                then if G.reverse 
                     then rev (s :: acc)
                     else (s :: acc) (* initial *)
                else
                    let
                        val i = indexof pred
                        val l = Array.sub(table, i)
                        fun go nil = raise Search "solve error!"
                          | go ({me=me', pred=pred', ...}::rest) =
                            if G.eq (me', pred)
                            then (* found it *)
                                solution (me' :: acc) (me', pred')
                            else go rest
                    in
                        go l
                    end
         
            fun one ((s, prev), d) =
                let
                    val sol = solution nil (s, prev)
                in
                    print ("--------------------------------------------------\n");
                    print ("            DISTANCE " ^ Int.toString d ^ ": (code " ^
                           G.tocode s ^ ")\n");
                    print ("--------------------------------------------------\n");
                    print (G.tostring s ^ "\n");
                    ListUtil.appi 
                       (fn (s, i) =>
                          let in
                              print ("--------------------------------------------------\n");
                              print ("Move " ^ Int.toString i ^ ":");
                              print (indent 4 ("\n" ^ G.tostring s) ^ "\n")
                          end) sol
                end

        in
            app one furthest;
            (* need one at the end for slide-mode *)
            print ("--------------------------------------------------\n")
        end

    (* then print out the whole table, if it is small enough *)
    val _ = 
        if nseen () < 500
        then
        let
            val _ = print "ALL STATES:\n"

            fun lookat {me, pred, dist, depth=d, ...} =
                let (* val d = depth 0 (me, pred) *)
                in
                    print ("board " ^ G.tocode me ^ " (pred " ^ G.tocode pred ^ " (dist" ^
                           Int.toString dist ^ ")) depth " ^ 
                           Int.toString d ^ "\n");
                    print (G.tostring me ^ "\n")
                end
        in
            Array.app (app lookat) table
        end
    else ()

end

exception Parse of string

val WIDTH = 16
val HEIGHT = 16
val WIDTHR = 18
val HEIGHTR = 18

(* bit masks for directions *)
val UP = 0w1 : Word32.word
val DOWN = 0w2 : Word32.word
val LEFT = 0w4 : Word32.word
val RIGHT = 0w8 : Word32.word

(* PERF maybe array? *)
fun dirchange 0w1 = ~WIDTHR
  | dirchange 0w2 = WIDTHR
  | dirchange 0w4 = ~1
  | dirchange 0w8 = 1
  | dirchange _ = 0

(* parses a standard 16x16 board. return any extra markers
   found in a list with their indices. *)
fun parse_walls filename =
    let
        val f = TextIO.openIn filename

        fun readuntil c =
            case TextIO.input1 f of
                NONE => raise Parse ("eof looking for chr " ^ 
                                       Int.toString (ord c))
              | SOME cc => if c = cc then ()
                           else readuntil c
            
        val board = Array.array(WIDTHR * HEIGHTR, 0w0 : Word32.word)

        val ents = ref (nil : (char * int) list)

        fun readtop y =
            let in
                Util.for 0 (WIDTH - 1)
                (fn x =>
                 (* read + *)
                 let val above = y * WIDTHR + (x + 1)
                     val below = (y + 1) * WIDTHR + (x + 1)
                 in
                     ignore (TextIO.input1 f);
                     case TextIO.input1 f of
                         SOME #"-" => 
                         let in
                             Array.update(board,
                                          above,
                                          Word32.orb
                                          (Array.sub(board, above),
                                           DOWN));
                             Array.update(board,
                                          below,
                                          Word32.orb
                                          (Array.sub(board, below),
                                           UP))
                         end
                       | SOME #" " => ()
                       (* XXX support v ^ *)
                       | SOME c => raise Parse ("bad char " ^ implode [c])
                       | NONE => raise Parse "eof"
                 end);
                readuntil #"\n"
            end handle e => raise e

        fun readmid y =
            let in
                if TextIO.input1 f = SOME #"|" 
                then ()
                else raise Parse "every row should start with |";

                Array.update(board, (y + 1) * WIDTHR,
                             Word32.orb(Array.sub(board, (y + 1) * WIDTHR),
                                        RIGHT));

                Array.update(board, (y + 1) * WIDTHR + 1,
                             Word32.orb(Array.sub(board, (y + 1) * WIDTHR + 1),
                                        LEFT));


                Util.for 0 (WIDTH - 1)
                (fn x =>
                 let
                     val left = (y + 1) * WIDTHR + (x + 1)
                     val right = (y + 1) * WIDTHR + (x + 2)
                 in
                     (* print ("reading for cell " ^
                        Int.toString left ^ "/" ^
                        Int.toString right ^ "\n"); *)
                     (* entity cell *)
                     (case TextIO.input1 f of
                          SOME #" " => ()
                        | SOME c => ents := (c, left) :: !ents
                        | NONE => raise Parse "eof(cell)");

                     (case TextIO.input1 f of
                          SOME #"|" => 
                              let in
                                  Array.update(board,
                                               left,
                                               Word32.orb
                                               (Array.sub(board, left),
                                                RIGHT));
                                  Array.update(board,
                                               right,
                                               Word32.orb
                                               (Array.sub(board, right),
                                                LEFT))
                              end
                        | SOME #" " => ()
                        | SOME c => raise Parse ("bad midchar " ^
                                                   implode [c])
                        | NONE => raise Parse "eof(mid)")
                 end);
                readuntil #"\n"
            end handle e => raise e
    in
       Util.for 0 (HEIGHT - 1)
          (fn y =>
           let in
               readtop y;
               readmid y
           end);

       (* and one more row on the bottom *)
       readtop HEIGHT;

       TextIO.closeIn f;

       {board = board, ents = !ents}
    end 

(* retract five yellow blocks (approximate; doesn't take player
   maneuvering into account) *)
functor Yellow5AllRetract(val filename : string) :> GAME =
struct

    val findshortest = false
    fun accepting _ = true
    val reverse = true

    (* always distinct and sorted *)
    type state = { y1 : Word9.word,
                   y2 : Word9.word,
                   y3 : Word9.word,
                   y4 : Word9.word,
                   y5 : Word9.word }

    fun sort_state { y1, y2, y3, y4, y5 } =
        let val l = [y1, y2, y3, y4, y5]

            val l = ListUtil.sort Word9.compare l
        in
            case l of
                [y1, y2, y3, y4, y5] => {y1=y1, y2=y2, y3=y3, y4=y4, y5=y5}
              | _ => raise Match (* impossible *)
        end


    fun merge y1 {y2, y3, y4, y5} =
      (if y1 <= y2
       then let val m2_y1 = y1
            in let val m2_y2 = y2
      in let val m2_y3 = y3
      in let val m2_y4 = y4
      in let val m2_y5 = y5
      in {y1 = m2_y1, y2 = m2_y2, y3 = m2_y3, y4 = m2_y4, y5 = m2_y5}
      end
      end
      end
      end
            end
       else let val m2_y1 = y2
            in (if y1 <= y3
       then let val m2_y2 = y1
            in let val m2_y3 = y3
      in let val m2_y4 = y4
      in let val m2_y5 = y5
      in {y1 = m2_y1, y2 = m2_y2, y3 = m2_y3, y4 = m2_y4, y5 = m2_y5}
      end
      end
      end
            end
       else let val m2_y2 = y3
            in (if y1 <= y4
       then let val m2_y3 = y1
            in let val m2_y4 = y4
      in let val m2_y5 = y5
      in {y1 = m2_y1, y2 = m2_y2, y3 = m2_y3, y4 = m2_y4, y5 = m2_y5}
      end
      end
            end
       else let val m2_y3 = y4
            in (if y1 <= y5
       then let val m2_y4 = y1
            in let val m2_y5 = y5
      in {y1 = m2_y1, y2 = m2_y2, y3 = m2_y3, y4 = m2_y4, y5 = m2_y5}
      end
            end
       else let val m2_y4 = y5
            in let val m2_y5 = y1
      in {y1 = m2_y1, y2 = m2_y2, y3 = m2_y3, y4 = m2_y4, y5 = m2_y5}
      end
            end)
            end)
            end)
            end)

    val xorb = Word32.xorb
    infix xorb
    fun hash {y1, y2, y3, y4, y5} =
        (* PERF: note, toLarge is apparently slower in current mlton *)
        Word32.<< (Word32.fromInt (Word9.toInt y1), 0w0)  xorb
        Word32.<< (Word32.fromInt (Word9.toInt y2), 0w7)  xorb
        Word32.<< (Word32.fromInt (Word9.toInt y3), 0w14)  xorb
        Word32.<< (Word32.fromInt (Word9.toInt y4), 0w21)  xorb
        Word32.<< (Word32.fromInt (Word9.toInt y5), 0w28)

    val eq = op=

       
    val {board, initial} =
        (* parse f *)
        let
            val {board, ents} = parse_walls filename

            fun indexof c =
                case ListUtil.Alist.find op= ents c of
                    SOME i => Word9.fromInt i
                  | NONE => raise Parse ("didn't find ent " ^ implode [c])


            val blocks = sort_state {y1 = indexof #"1", 
                                     y2 = indexof #"2", 
                                     y3 = indexof #"3", 
                                     y4 = indexof #"4", 
                                     y5 = indexof #"5"}
        in
            { board = board, 
              initial = blocks }
        end handle Parse s =>
            let in
                print ("parse error: " ^ s ^ "\n");
                raise Parse s
            end

    fun app_successors (f, {y1, y2, y3, y4, y5} : state) =
        let

            val ti = Word9.toInt

            (* give it one of the five, and the rest *)
            fun try_one y (y1, y2, y3, y4) =
                let
                    (* val _ = print ("try " ^ Int.toString y ^ "\n") *)

                    (* dir is the direction the block was kicked,
                       so opp is the dir we are 'retracting in' *)
                    fun go dir opp =
                        let
                            
                            (* might be nice if we could use board for
                               this, but it's hard to maintain that it
                               is in the correct state *)

                            fun blockat dest =
                                dest = y1 orelse
                                dest = y2 orelse
                                dest = y3 orelse
                                dest = y4


                            (* in order to retract, must have something
                               blocking us in 'dir' to hit *)
                            fun blocked idx d opp =
                                let val dest = idx + dirchange d
                                    val w = Array.sub(board, dest)
                                    (*
                                    val _ = print ("  -- board[" ^ Int.toString dest ^ "]: " ^ 
                                                   Word32.toString w ^ "\n")
                                    *)
                                in
                                    Word32.andb(Array.sub(board, dest), opp) > 0w0
                                    orelse blockat dest
                                end

                            (* any free spot that this block could
                               have been pushed from (along the direction specified)
                               now counts as a valid predecessor *)

                            (*       + + + + +
                              opp <- |      i| -> dir
                                     + + + + + 

                              reasons to stop:
                                     + + + + +
                                     |  X i  |
                                     + + + + +
                                .. block before us.

                                     + + + + +
                                     |   |i  |
                                     + + + + +
                                        -> (blocked)
                                .. wall before us.

                              *)

                            and retracting idx dir opp =
                                (* retract at least once first, since when
                                   this is called, we have already visited
                                   that state *)
                                let
                                    val newi = idx + dirchange opp
                                    val newi2 = newi + dirchange opp
                                        
                                    (*
                                    val _ = print ("  (idx = " ^ Int.toString idx ^
                                                   " dir = " ^ Word32.toString dir ^ 
                                                   " newi = " ^ Int.toString newi ^ ")\n");
                                    *)
                                in
                                    (* PERF: this could be more efficient, since
                                       these tests overlap between iterations.
                                       make a loop invariant, establish it, and use it *)
                                    if blockat newi orelse blockat newi2 
                                       orelse blocked newi dir opp
                                       orelse blocked newi2 dir opp
                                    then () (* done -- can't push from here *)
                                    else let
                                             val fi = Word9.fromInt
                                             val news = merge (fi newi) { y2 = fi y1, 
                                                                          y3 = fi y2,
                                                                          y4 = fi y3, 
                                                                          y5 = fi y4}
                                         in
                                             f (news, 1);
                                             retracting newi dir opp
                                         end
                                end

                        in
                            if blocked y dir opp
                            then retracting y dir opp
                            else () (* print ("  .. not blocked " ^ Word32.toString dir ^ "\n") *)
                        end
                        
                in
                    go UP DOWN;
                    go DOWN UP;
                    go LEFT RIGHT;
                    go RIGHT LEFT
                end

            (* can also pull it out of the target, if desired *)
        (* XX no target *)
            fun try_one' y r = try_one y r

            val y1 = ti y1
            val y2 = ti y2
            val y3 = ti y3
            val y4 = ti y4
            val y5 = ti y5
        in
            try_one' y1 (y2, y3, y4, y5);
            try_one' y2 (y1, y3, y4, y5);
            try_one' y3 (y1, y2, y4, y5);
            try_one' y4 (y1, y2, y3, y5);
            try_one' y5 (y1, y2, y3, y4)
        end

    fun tostring {y1, y2, y3, y4, y5} =
        let
            val e = map Word9.toInt [y1, y2, y3, y4, y5]
            val ll =
                List.tabulate 
                (HEIGHTR,
                 (fn y =>
                  List.tabulate
                  (WIDTHR,
                   (fn x => (Array.sub (board, y * WIDTHR + x), y * WIDTHR + x)))))
        in
            StringUtil.delimit "\n"
            (map 
             (fn l => 
              String.concat 
              ((map (fn (w, _) => "+" ^ (if Word32.andb(w, UP) > 0w0 then "-" else " ") ^ "+") l) @
               "\n" ::
               (map (fn (w, i) => (if Word32.andb(w, LEFT) > 0w0
                              then "|" else " ") ^ 
                              (if List.exists (fn z => z = i) e
                               then "X"
                               else " ") ^
                             (if Word32.andb(w, RIGHT) > 0w0
                              then "|" else " ")) l) @ "\n" ::
               (map (fn (w, _) => "+" ^ (if Word32.andb(w, DOWN) > 0w0 then "-" else " ") ^ "+") l))) ll)
        end

    fun tocode {y1, y2, y3, y4, y5} =
        let
            val e = map Word9.toString [y1, y2, y3, y4, y5]
        in
            StringUtil.delimit "," e
        end
end

functor Grey6Retract(val filename : string) :> GAME =
struct

    val findshortest = false
    fun accepting _ = true
    val reverse = true

    (* always distinct and sorted *)
    type state = { guy : Word9.word,
                   y1 : Word9.word,
                   y2 : Word9.word,
                   y3 : Word9.word,
                   y4 : Word9.word,
                   y5 : Word9.word,
                   y6 : Word9.word
                   }

    (* this can be slow because it is only used in initialization *)
    fun sort_state { guy, y1, y2, y3, y4, y5, y6 } =
        let val l = [y1, y2, y3, y4, y5, y6]

            val l = ListUtil.sort Word9.compare l
        in
            case l of
                [y1, y2, y3, y4, y5, y6] => {guy=guy, y1=y1, y2=y2, y3=y3, y4=y4, y5=y5, y6=y6}
              | _ => raise Match (* impossible *)
        end
        
    (* generated by gensort.sml *)
    fun merge guy y r =
        let val {y1, y2, y3, y4, y5, y6} = merge' y r
        in
            {guy=guy, y1=y1, y2=y2, y3=y3, y4=y4, y5=y5, y6=y6}
        end

    and merge' y1 {y2 : Word9.word, y3, y4, y5, y6} =
(if y1 <= y2
 then let val m1_y1 = y1
      in let val m1_y2 = y2
in let val m1_y3 = y3
in let val m1_y4 = y4
in let val m1_y5 = y5
in let val m1_y6 = y6
in {y1 = m1_y1, y2 = m1_y2, y3 = m1_y3, y4 = m1_y4, y5 = m1_y5, y6 = m1_y6}
end
end
end
end
end
      end
 else let val m1_y1 = y2
      in (if y1 <= y3
 then let val m1_y2 = y1
      in let val m1_y3 = y3
in let val m1_y4 = y4
in let val m1_y5 = y5
in let val m1_y6 = y6
in {y1 = m1_y1, y2 = m1_y2, y3 = m1_y3, y4 = m1_y4, y5 = m1_y5, y6 = m1_y6}
end
end
end
end
      end
 else let val m1_y2 = y3
      in (if y1 <= y4
 then let val m1_y3 = y1
      in let val m1_y4 = y4
in let val m1_y5 = y5
in let val m1_y6 = y6
in {y1 = m1_y1, y2 = m1_y2, y3 = m1_y3, y4 = m1_y4, y5 = m1_y5, y6 = m1_y6}
end
end
end
      end
 else let val m1_y3 = y4
      in (if y1 <= y5
 then let val m1_y4 = y1
      in let val m1_y5 = y5
in let val m1_y6 = y6
in {y1 = m1_y1, y2 = m1_y2, y3 = m1_y3, y4 = m1_y4, y5 = m1_y5, y6 = m1_y6}
end
end
      end
 else let val m1_y4 = y5
      in (if y1 <= y6
 then let val m1_y5 = y1
      in let val m1_y6 = y6
in {y1 = m1_y1, y2 = m1_y2, y3 = m1_y3, y4 = m1_y4, y5 = m1_y5, y6 = m1_y6}
end
      end
 else let val m1_y5 = y6
      in let val m1_y6 = y1
in {y1 = m1_y1, y2 = m1_y2, y3 = m1_y3, y4 = m1_y4, y5 = m1_y5, y6 = m1_y6}
end
      end)
      end)
      end)
      end)
      end)

    (* only player can go here *)
    val ROUGH = 0w16

    val xorb = Word32.xorb
    infix xorb
    fun hash {guy, y1, y2, y3, y4, y5, y6} =
        (* PERF: note, toLarge is apparently slower in current mlton *)
        Word32.fromInt (Word9.toInt guy)  xorb
        Word32.<< (Word32.fromInt (Word9.toInt y1), 0w1)  xorb
        Word32.<< (Word32.fromInt (Word9.toInt y2), 0w8)  xorb
        Word32.<< (Word32.fromInt (Word9.toInt y3), 0w15)  xorb
        Word32.<< (Word32.fromInt (Word9.toInt y4), 0w22)  xorb
        Word32.<< (Word32.fromInt (Word9.toInt y5), 0w29)  xorb
        Word32.<< (Word32.fromInt (Word9.toInt y6), 0w18)  

    val eq = op=

       
    val {board, initial} =
        (* parse f *)
        let
            val {board, ents} = parse_walls filename

            fun indexof c =
                case ListUtil.Alist.find op= ents c of
                    SOME i => Word9.fromInt i
                  | NONE => raise Parse ("didn't find ent " ^ implode [c])


            val initial = sort_state { guy = indexof #"G",
                                      y1 = indexof #"1", 
                                      y2 = indexof #"2", 
                                      y3 = indexof #"3", 
                                      y4 = indexof #"4", 
                                      y5 = indexof #"5", 
                                      y6 = indexof #"6" }

            fun setrough (#":", i) =
                Array.update(board, i, Word32.orb(Array.sub(board, i),
                                                  ROUGH))
              | setrough _ = ()
        in
            app setrough ents;

            { board = board, 
              initial = initial }
        end handle Parse s =>
            let in
                print ("parse error: " ^ s ^ "\n");
                raise Parse s
            end

    fun app_successors (f, {guy, y1, y2, y3, y4, y5, y6} : state) =
        let

            (* can move the guy in one of four directions. *)

            val ti = Word9.toInt

            val guy = ti guy
            val y1 = ti y1
            val y2 = ti y2
            val y3 = ti y3
            val y4 = ti y4
            val y5 = ti y5
            val y6 = ti y6

            fun blockat dest =
                dest = y1 orelse
                dest = y2 orelse
                dest = y3 orelse
                dest = y4 orelse
                dest = y5 orelse
                dest = y6

            (* retracting a move in direction 'dir' 

               <- opp
                -+-+-+
                  G X|
                -+-+-+
                  dir->

               retracts to:

                -+-+-+           -+-+-+
                G   X|     or    G X  |
                -+-+-+           -+-+-+
               

               *)
            fun go dir opp =
                let
                    val guynew = guy + dirchange opp

                    (* here, blocked means that we wouldn't
                       have been able to walk into 'idx' from
                       d. We also assume we don't ever start
                       on walls or blocks. *)
                    val blocked =
                        let 
                        in
                            (*
                            print (" guy: " ^ Int.toString guy ^ "=" ^ Word32.toString(Array.sub(board,guy)) ^
                                   " dir: " ^ Word32.toString dir ^ 
                                   " opp: " ^ Word32.toString opp ^
                                   " new: " ^ Int.toString guynew ^ "=" ^ Word32.toString(Array.sub(board,guynew)) ^
                                   " cond1: " ^ Bool.toString (Word32.andb(Array.sub(board, guy), dir) > 0w0)^
                                   " cond2: " ^ Bool.toString (Word32.andb(Array.sub(board, guynew), opp) > 0w0) ^
                                   " cond3: " ^ Bool.toString (blockat guynew) ^
                                   " cond4: " ^ Bool.toString (blockat guy) ^
                                   "\n");
                            *)

                            Word32.andb(Array.sub(board, guy), opp) > 0w0 orelse
                            Word32.andb(Array.sub(board, guynew), dir) > 0w0 orelse
                            blockat guynew
                        end

                in
                    if blocked
                    then ()
                    else (* retraction possible (just moving). *)
                        let 
                            val fi = Word9.fromInt

                            (* XXX consider this edge 0 -- or small? *)
                            val _ = f ({ guy = fi guynew, y1= fi y1, y2= fi y2, 
                                        y3= fi y3, y4= fi y4, y5= fi y5, y6= fi y6 }, 1)

                            val bi = guy + dirchange dir

                            (* what about bringing a block with us? *)
                            fun perhapsretract yb (yy1, yy2, yy3, yy4, yy5) =
                                if yb = bi
                                then f (merge (fi guynew) (fi guy)
                                         {y2 = fi yy1, y3= fi yy2, y4= fi yy3, y5= fi yy4,
                                          y6 = fi yy5},
                                        100)
                                else ()
                        in
                            (* can't be on rough. *)
                            if Word32.andb(Array.sub(board, guy), ROUGH) = 0w0
                                (* assume no wall between us -- we support only solid
                                   walls, I guess *)
                            then
                                let in
                                    (* and there must be a block to retract *)
                                    perhapsretract y1 (y2, y3, y4, y5, y6);
                                    perhapsretract y2 (y1, y3, y4, y5, y6);
                                    perhapsretract y3 (y1, y2, y4, y5, y6);
                                    perhapsretract y4 (y1, y2, y3, y5, y6);
                                    perhapsretract y5 (y1, y2, y3, y4, y6);
                                    perhapsretract y6 (y1, y2, y3, y4, y5)
                                end
                            else ()
                        end
                end
                    
        in
            go UP DOWN;
            go DOWN UP;
            go LEFT RIGHT;
            go RIGHT LEFT
        end
    

    fun tostring {guy, y1, y2, y3, y4, y5, y6} =
        let
            val guy = Word9.toInt guy
            val e = map Word9.toInt [y1, y2, y3, y4, y5, y6]
            val ll =
                List.tabulate 
                (HEIGHTR,
                 (fn y =>
                  List.tabulate
                  (WIDTHR,
                   (fn x => (Array.sub (board, y * WIDTHR + x), y * WIDTHR + x)))))
        in
            StringUtil.delimit "\n"
            (map 
             (fn l => 
              String.concat 
              ((map (fn (w, _) => "+" ^ (if Word32.andb(w, UP) > 0w0 then "-" else " ") ^ "+") l) @
               "\n" ::
               (map (fn (w, i) => (if Word32.andb(w, LEFT) > 0w0
                              then "|" else " ") ^ 
                              (if List.exists (fn z => z = i) e
                               then "X"
                               else if i = guy
                                    then "G"
                                    else if Word32.andb(w, ROUGH) > 0w0
                                         then ":"
                                         else " ") ^
                             (if Word32.andb(w, RIGHT) > 0w0
                              then "|" else " ")) l) @ "\n" ::
               (map (fn (w, _) => "+" ^ (if Word32.andb(w, DOWN) > 0w0 then "-" else " ") ^ "+") l))) ll)
        end

    fun tocode {guy, y1, y2, y3, y4, y5, y6} =
        let
            val e = map Word9.toString [guy, y1, y2, y3, y4, y5, y6]
        in
            StringUtil.delimit "," e
        end
end

(* "Forward" with (subset of) standard escape rules.
   Here we are not making a determination of "interesting"
   vs "uninteresting" features of the level, so a state is
   an actual board---this means that in practice we are limited
   to much smaller boards. (but we can generate much more interesting
   levels!) *)

(* Boards are limited to 16x16, but they can be cut short
   so that we don't have to allocate useless parts of the level. *)

functor EscapeFwd(val filename : string) :> GAME =
struct

  val findshortest = true
  val reverse = false

  datatype tile = FLOOR | GREEN | BLUE | GREY | ROUGH | GOLD | EXIT | LASER

  type state = { guy : Word8.word,
                 board : tile Vector.vector }
                 (* could add bots, etc. *)

  val rols = Vector.fromList [0w3, 0w5, 0w7, 0w11, 0w13, 0w17, 0w19, 0w23 : Word.word]

  (* this doesn't even depend on the whole board,
     since hashing would be very expensive *)
  fun hash { guy, board } =
    let 
      fun tw FLOOR  = 0wx11335577
        | tw GREEN  = 0wx98765432
        | tw BLUE   = 0wxFFAACCDD
        | tw GREY   = 0wx73737373
        | tw ROUGH  = 0wx18274799
        | tw GOLD   = 0wxA0C2B3D9
        | tw EXIT   = 0wx3F7F5A90
        | tw LASER  = 0wx789ABCDE

      val s = Vector.length board
        
      fun rol (w, n) = Word32.orb(Word32.<< (w, n),
                                  Word32.>> (w, 0w32 - n))

      val t = 0wx00000000
      val i = Word8.toInt guy

      fun go (t, _) 0 = t
        | go (t, i) n =
        let
          val b = tw (Vector.sub(board,(i+Word32.toInt(Word32.andb(t,0wxFFFFFF))) mod s))
          val t = Word32.xorb(rol(t, Vector.sub(rols, i mod 8)), b)
          val i = i + 1
        in
          go (t, i) (n - 1)
        end
    in
      (* higher number of iterations gives better quality hash,
         but this whole thing is pointless if n gets on the order of
         the size of the board. *)
      Word32.xorb(Word32.fromInt i,
                  go (t, i) 24)
    end

  val eq = op=

  val (width, initial) =
    let

      val f = TextIO.openIn filename
        
      fun readuntil c =
        case TextIO.input1 f of
          NONE => raise Parse ("eof looking for chr " ^ 
                               Int.toString (ord c))
        | SOME cc => if c = cc then ()
                     else readuntil c


      fun uchar #"#" = BLUE
        | uchar #" " = FLOOR
        | uchar #":" = ROUGH
        | uchar #"*" = GOLD
        | uchar #"@" = GREY
        | uchar #"E" = EXIT
        | uchar #"%" = GREEN
        | uchar #"G" = FLOOR
        | uchar #"L" = LASER
        | uchar c = raise Parse ("Unexpected char: " ^ Char.toString c)

      val g = ref NONE

      (* width, once we figure it out *)
      val w = ref ~1

      val arr = Array.array(16 * 16, FLOOR)
        
      fun parse i =
        (case TextIO.input1 f of
           NONE => raise Parse ("Unexpected EOF: require ^ at end")
         | SOME #"<" => 
             (* end row *)
             let in
               (case !w of
                  ~1 => w := i
                | x  => if x mod !w <> 0 
                        then raise Parse ("Uneven rows")
                        else ());
               readuntil #"\n";
               parse i
             end
         | SOME #"^" =>
             (* end everything *)
             let in
               (* height *)
               i div !w
             end
         | SOME c => 
                let in
                  if c = #"G"
                  then g := SOME i
                  else ();
                  Array.update(arr, i, uchar c);
                  parse (i + 1)
                end)
             
      val h = parse 0
    in
      
      TextIO.closeIn f;
      (!w,
       { guy = 
         (case !g of
            NONE => raise Parse ("There is no guy (G)")
          | SOME i => Word8.fromInt i),
         board = Vector.tabulate (!w * h,
                                  (fn i => Array.sub(arr, i))) })
    end handle (e as Parse s) => (print ("parse error: " ^ s ^ "\n");
                                  raise e)

  (* require guy to be on accepting state! *)
  fun accepting { guy, board } =
    let val gi = Word8.toInt guy
    in
      case Vector.sub(board, gi) of
        EXIT => true
      | _ => false
    end


  fun app_successors (f, { guy, board }) =
    let

      fun index (x, y) = y * width + x
      fun whereis i = (i mod width, i div width)
      fun boardat (x, y) = Vector.sub(board, index(x, y))

      val height = Vector.length board div width

      datatype dir = Up | Down | Left | Right

      fun travel Up (_, 0) = NONE
        | travel Up (x, y) = SOME (x, y - 1)
        | travel Down (x, y) = if y < (height - 1)
                               then SOME (x, y + 1)
                               else NONE
        | travel Left (0, y) = NONE
        | travel Left (x, y) = SOME (x - 1, y)
        | travel Right (x, y) = if x < (width - 1)
                                then SOME (x + 1, y)
                                else NONE
        
      val (gx, gy) = whereis (Word8.toInt guy)


      val COST_WALK = 1
      val COST_GREY = 1
      val COST_GOLD = 1

      (* only call f if the resulting board does not kill
         the player *)
      val f = 
        (* nb, now working with new board and guy, so
           be careful *)
        (fn stuff as ({ guy, board }, cost) =>
         let
           fun dead d =
             let
               fun de (x, y) =
                 case travel d (x, y) of
                   NONE => false
                 | SOME (nx, ny) => 
                     (case Vector.sub(board, index(nx, ny)) of
                        LASER => true
                      | FLOOR => de (nx, ny)
                      | ROUGH => de (nx, ny)
                      (* anything else blocks laser *)
                      | _ => false) 

               val (x, y) = whereis (Word8.toInt guy)
             in
               de (x, y)
             end
         in
           if dead Up orelse
              dead Down orelse
              dead Left orelse
              dead Right
           then ()
           else f stuff
         end)

      (* XXX parameterize cost *)
      fun walk (tx, ty) =
        (* can walk there *)
        f ({ guy = Word8.fromInt (index (tx, ty)), 
             board = board }, COST_WALK)

      fun move d =
        case travel d (gx, gy) of
          NONE => () (* edge of level *)
        | SOME (tx, ty) => 
            (case boardat (tx, ty) of
               FLOOR => walk (tx, ty)
             | EXIT => walk (tx, ty)
             | ROUGH => walk (tx, ty)

             | GOLD =>
                 let
                   fun going (x, y) =
                     case travel d (x, y) of
                       NONE => (x, y)
                     | SOME r => 
                         (case boardat r of
                            FLOOR => going r
                          | _ => (x, y))

                   val (fx, fy) = going (tx, ty)

                   val dst = index (fx, fy)
                   val src = index (tx, ty)
                 in
                   if dst = src
                   then ()
                   else
                     (* slide *)
                     f ({ guy = guy,
                          board =
                          Vector.tabulate
                          (Vector.length board,
                           (fn i =>
                            if i = src then FLOOR
                            else if i = dst then GOLD
                                 else Vector.sub(board, i))) },
                        COST_GOLD)
                 end
             | GREY => 
             (* push only if dest is floor *)
                 (case travel d (tx, ty) of
                    NONE => ()
                  | SOME (dx, dy) =>
                      let 
                        val src = index (tx, ty)
                        val dst = index (dx, dy)
                      in
                        if boardat (dx, dy) = FLOOR
                        then
                          (* push *)
                          f ({ guy = Word8.fromInt (index (tx, ty)),
                               board = 
                               Vector.tabulate 
                               (Vector.length board,
                                (fn i =>
                                 if i = src then FLOOR
                                 else if i = dst then GREY
                                      else Vector.sub(board, i))) }, 
                             COST_GREY)
                        else ()
                      end)

             | GREEN => 
             (* like grey, but becomes blue *)
                 (case travel d (tx, ty) of
                    NONE => ()
                  | SOME (dx, dy) =>
                      let 
                        val src = index (tx, ty)
                        val dst = index (dx, dy)
                      in
                        if boardat (dx, dy) = FLOOR
                        then
                          (* push *)
                          f ({ guy = Word8.fromInt (index (tx, ty)),
                               board = 
                               Vector.tabulate 
                               (Vector.length board,
                                (fn i =>
                                 if i = src then FLOOR
                                 else if i = dst then BLUE
                                      else Vector.sub(board, i))) }, 
                             COST_GREY)
                        else ()
                      end)

             (* should be dead, anyway *)
             | LASER => ()
             | BLUE => ())

    in
      (* don't move off exit, since that can only be done on
         the first move and the parser doesn't allow such an
         initial state to be parsed. ;) *)
      case boardat (gx, gy) of
        EXIT => ()
      | _ => 
          let in
            (* basically, the only moves are the four cardinal
               directions. *)
            move Up;
            move Down;
            move Left;
            move Right
          end
    end

  fun tostring {guy, board} =
    let 
      val topbot = CharVector.tabulate(width + 2, fn _ => #"#")
    in
    StringUtil.delimit "\n"
    (topbot ::
     List.tabulate
     (Vector.length board div width,
      (fn y =>
       "#" ^
       CharVector.tabulate
       (width,
        (fn x =>
         let val i = y * width + x
         in
           if i = Word8.toInt guy
           then #"G"
           else 
             case Vector.sub(board, i) of
               BLUE => #"#"
             | ROUGH => #":"
             | FLOOR => #" "
             | GREY => #"@"
             | GOLD => #"*"
             | EXIT => #"E"
             | GREEN => #"%"
             | LASER => #"L"
         end)) ^ "#")) @ [topbot])
    end

  (* no sensible code here... *)
  fun tocode s = Word32.toString (hash s)

end

(*
structure Y = Yellow5AllRetract(val filename = "board.txt")
structure R = Search(Y)
*)

(*
structure G = Grey6Retract(val filename = "soko.txt")
structure R = Search(G)
*)

structure E = EscapeFwd(val filename = "esc.txt")
val _ = print (E.tostring E.initial ^ "\n")
structure R = Search(E)