
(* PERF: use an array now that nid = int *)
structure Graph : GRAPH = 
struct

  exception Graph of string

  (* small integer structure; used to get better memory usage.
     MLton has all IntN for N in (2..32, 64).
     we will get Overflow if the number of nodes is greater than
     2*(N-1)-1, so give a little slack here (16 should be good) *)
  structure SI = Int16

  type nid = Types.nid
  type node = Types.node
  type edge_tag = Types.edge_tag
  type edge = Types.edge
  type ptype = Types.ptype

  datatype direction = Forward | Backward (* used in foot *)
  datatype transport = Only of ptype
                     | Any of ptype (* ptype gives your current mode *)
 
  structure NM = SplayMapFn (struct 
                                 type ord_key = nid
                                 val compare = Int.compare
                             end)

  structure SM = SplayMapFn (struct
                                 type ord_key = string
                                 val compare = String.compare
                             end)

  (* represents graph structure, but not augmented
     with precalculations on it *)
  type pregraph = 
     { 
       name_to_nid_map : nid SM.map,
       nid_to_name_map : string Vector.vector,

       (* PERF could also be vector *)
       map : 
       { out_edges : (nid * edge_tag) list,
         in_edges : (nid * edge_tag) list} NM.map,
      
       nelements : int,

       nodetype : Types.node_tag Vector.vector,
       hq : int,
       banks : int list,
       robberstart : int }
       
  (* graph is represented by a collection of nodes (the domain
     of this map); for each one there is a list of adjacent
     nodes (the "edges") and a map giving the distance of each
     node by foot. *)

  type step = (nid * ptype)
  type path = step list

  type graph = 
      { pregraph : pregraph,
        paths : (transport -> SI.int Vector.vector) Vector.vector}

  val dummy = { pregraph = {
                            name_to_nid_map = SM.empty,
                            nid_to_name_map = Vector.fromList nil,
                            map = NM.empty,
                            nelements = 0,
                            nodetype = Vector.fromList nil,
                            hq = 0,
                            banks = nil,
                            robberstart = 0 },
                paths = Vector.fromList nil} : graph

  type biedges = {in_edges : (nid * edge_tag) list,
                  out_edges : (nid * edge_tag) list}
                 
  fun pl f (x::xs) = (print (f x); print " "; pl f xs)
    | pl f nil = print "\n"

  (* PERF might want to tune this since adj is a VERY popular function *)
  fun union (xs, ys) =
      let 
          (* XXX exists *)
          fun contains x xs = (case List.find (fn x' => x = x') xs
                                   of SOME _ => true
                                 | NONE => false)
      in
          List.foldl (fn (y, acc) => 
                      if contains y acc then acc
                      else y::acc) xs ys
      end

  fun topregraph (nodes, edges) =
      let (* Turn Foot edges into Both edges on the first pass. *)
          fun tt et = case et of Types.Foot => Types.Both
                                | Types.Car => Types.Car
                                | Types.Both => 
                                  raise Graph "didn't expect Both edge"
          fun out_adj from =
              List.mapPartial (fn (from',to,edge_tag) =>
                                  if (from=from') then SOME (to,tt edge_tag)
                                  else NONE)
                              edges
          fun in_adj to =
              List.mapPartial (fn (from,to',edge_tag) =>
                                  if (to=to') then SOME (from,tt edge_tag)
                                  else NONE)
                              edges
          val result = 
              foldr (fn (x, adjmap) =>
                        NM.insert (adjmap, x, {out_edges=out_adj x,
                                               in_edges=in_adj x}))
                    NM.empty (map (fn (nid,_,_,_) => nid) nodes)

          (* Then put in Foot edges in the backward direction. *)
          fun out_adj_back from =
              List.mapPartial (fn (from',to,edge_tag) =>
                                  if from=to andalso edge_tag = Types.Foot 
                                  then SOME (from',Types.Foot)
                                  else NONE)
                              edges
          fun in_adj_back to =
              List.mapPartial (fn (from,to',edge_tag) =>
                                  if from=to andalso edge_tag = Types.Foot
                                  then SOME (to',Types.Foot)
                                  else NONE)
                              edges

          val result_back =
              foldr (fn (x, adjmap) =>
                        NM.insert (adjmap, x, {out_edges=out_adj_back x,
                                               in_edges=in_adj_back x}))
                    result (map (fn (nid,_,_,_) => nid) nodes)

          fun my_union (e1 : biedges, e2 : biedges) =
              {in_edges=(union (#in_edges e1, #in_edges e2)),
               out_edges=(union (#out_edges e1, #out_edges e2))}
      in 
        NM.unionWith my_union (result, result_back)
      end

  fun buildnames (nodes, edges) =
      let 
        val names = Array.array(length nodes, "")
        val (name_to_nid_map, _) =
              foldr (fn ((name,_,_,_), (name_to_nid_map, n)) => 
                     let in
                       Array.update(names, n, name);
                       (SM.insert (name_to_nid_map, name, n), n + 1)
                     end)
                    (SM.empty, 0) nodes
      in
        (name_to_nid_map, VectorUtil.atov names)
      end

  fun name_to_nid ({pregraph = {name_to_nid_map, ...}, ...} : graph) name = 
      case SM.find (name_to_nid_map, name) of
        NONE => raise Graph ("unknown node name " ^ name)
      | SOME nid => nid

  fun nid_to_name ({pregraph = {nid_to_name_map, ...}, ...} : graph) nid = 
      Vector.sub(nid_to_name_map, nid) handle e => raise e

  fun any_filter (pg : pregraph) nd pt (_, etag) = 
      (case (pt, etag) 
        of (Types.CopFoot, Types.Foot) => true
         | (Types.CopCar, Types.Car) => true
         | (Types.Robber, _) => raise Graph "didn't expect robber for Any"
         | (_, Types.Both) => true
         (* cops can switch if when leaving hq *)
         | (Types.CopFoot, Types.Car) => nd = #hq pg
         | (Types.CopCar, Types.Foot) => nd = #hq pg)


  fun adjacent_pre (pg as {map, ...} : pregraph) dir trans from = 
      let val edges = 
              case dir of Forward => (case NM.find (map, from) of
                                        SOME {out_edges, ...} => out_edges
                                      | NONE => raise Graph "bad nid/edges?")
                        | Backward => (case NM.find (map, from) of
                                         SOME {in_edges, ...} => in_edges
                                       | NONE => raise Graph "bad nid/edges?")
                                      
          val f = case trans
                   of Only Types.CopFoot =>
                      (fn (nd, etag) => (etag = Types.Both 
                                         orelse etag = Types.Foot))
                    | Only Types.CopCar => 
                      (fn (nd, etag) => etag = Types.Both
                                        orelse etag = Types.Car)
                    | Only Types.Robber => 
                      (fn (nd, etag) => (etag = Types.Both 
                                         orelse etag = Types.Foot))
                    | Any pt => any_filter pg from pt

          val edges = List.filter f edges

          fun etag_to_ptype ((nd, Types.Both), steps) =
              (case trans of Only pt => (nd, pt)::steps
                           | Any pt => if from = #hq pg 
                                       then (nd, Types.CopFoot)
                                            ::(nd, Types.CopCar)::steps
                                       else (nd, pt)::steps)
            | etag_to_ptype ((nd, Types.Foot), steps) =
              (case trans of Only pt => (nd, pt)::steps
                           | Any pt => (nd, Types.CopFoot)::steps)

            | etag_to_ptype ((nd, Types.Car), steps) =
              (case trans of Only pt => (nd, pt)::steps
                           | Any pt => (nd, Types.CopCar)::steps)

      in
        union (nil, List.foldl etag_to_ptype nil edges)
      end

  structure IH = HeapFn(type priority = int
                        val compare = Int.compare)

  fun dijkstra n src orig_trans (pregraph as {map, nelements, ...} : pregraph)
    : SI.int Vector.vector =
      let
          (* PERF -- should use arrays for all of this instead
             of maps *)
          (* PERF path computation is useless now (although mlton may
             be smart enough to figure that out *)

          (* place to hold hands *)
          val hands_foot = Array.array(nelements, NONE)
          val hands_car  = Array.array(nelements, NONE)

          (* current distances *)
          val dist_foot  = Array.array(nelements, 8888)
          val dist_car   = Array.array(nelements, 8887)
                               
          fun sethand n pt v =
              Array.update((case pt of
                                Types.CopCar => hands_car
                              | _ => hands_foot), n, v)

          fun gethand n pt =
              Array.sub((case pt of
                             Types.CopCar => hands_car
                           | _ => hands_foot), n)

          fun getdist n pt =
              Array.sub((case pt of
                             Types.CopCar => dist_car
                           | _ => dist_foot), n)

          fun setdist n pt v =
              Array.update((case pt of
                                Types.CopCar => dist_car
                              | _ => dist_foot), n, v)

          val h = IH.empty ();
          val pt = case orig_trans of Only pt => pt | Any pt => pt

          (* put the source in the frontier *)
          val _ = sethand src pt (SOME (IH.insert h 0 (src, pt)))
          val _ = setdist src pt 0

          (* now keep going until the frontier is exhausted *)
          fun expand () =
              case IH.min h of
                NONE => () (* done *)
              | SOME (dist, (node, cur_pt)) => 
                let val new_dist = dist + 1
                    (* don't look at car edges *)
                    fun neighbor (n, new_pt) =
                        (* enqueue it if it's our first, or update the
                           distance if we found a shorter path to it *)
                        (case gethand n new_pt of
                             NONE =>
                                 (* not yet seen *)
                                 let in
                                     sethand n new_pt
                                       (SOME (IH.insert h new_dist (n, new_pt)));
                                     setdist n new_pt new_dist
                                 end
                           | SOME ha =>
                                 (if IH.valid ha 
                                  then
                                      (* still in frontier, so update
                                         it if the new distance is better *)
                                      let 
                                          val (old_dist, (_, old_pt)) = 
                                              IH.get h ha
                                      in
                                          (* XXX is this right? it seems
                                             like we should have only one
                                             hand per node/transport item
                                             in the frontier. *)
                                          if old_dist > new_dist 
                                             andalso old_pt = new_pt
                                          then 
                                              let in
                                                  IH.adjust h ha new_dist;
                                                  setdist n old_pt new_dist
                                              end
                                          else ()
                                      end
                                  else ())) (* already found shortest path *)

                    (* If we are allow to switch, note that here. *)
                    val trans = 
                        case orig_trans of 
                            Only pt => Only pt
                          | Any _ => Any cur_pt
                                
                    val neighbors = adjacent_pre pregraph Forward trans node
                in
                  app neighbor neighbors;
                  expand ()
                end

          val _ = expand ()
      in
        Vector.tabulate
        (n,
         (fn me => 
          (* now read the appropriate distance using the requested
             transport. If we allow either kind of transport, then
             the distance is the minimum of foot/car *)
          case orig_trans of
              Only Types.CopCar => SI.fromInt (getdist me Types.CopCar)
            | Only _ => SI.fromInt (getdist me Types.CopFoot)
            | Any _ => SI.fromInt 
                        (Int.min
                         (getdist me Types.CopCar,
                          getdist me Types.CopFoot))))
      end

  fun msgnode_to_node graph (name, ntag, x, y) =
      (name_to_nid graph name, ntag, x, y)
  fun msgedge_to_edge graph (name1, name2, etag) =
      (name_to_nid graph name1, name_to_nid graph name2, etag)

  fun tograph (r as (orig_nodes, orig_edges)) =
      let
          val (name_to_nid_map, nid_to_name_map) = buildnames r

          (* XXX These next few functions are copied (almost) from above *)
          fun name_to_nid name = 
              case SM.find (name_to_nid_map, name) of
                NONE => raise Graph ("unknown node name " ^ name)
              | SOME nid => nid

          fun nid_to_name nid = 
            Vector.sub(nid_to_name_map, nid) handle e => raise e

          fun msgnode_to_node (name, ntag, x, y) =
              (name_to_nid name, ntag, x, y)
          fun msgedge_to_edge (name1, name2, etag) =
              (name_to_nid name1, name_to_nid name2, etag)

          val nodes = List.map msgnode_to_node orig_nodes
          val edges = List.map msgedge_to_edge orig_edges

          val n = List.length nodes

          (* they had better be dense, or this will not work *)
          val nos = map (fn (name, tag, _, _) =>
                         (name_to_nid name, tag : Types.node_tag)) orig_nodes
          val nos = ListUtil.sort 
                       (ListUtil.byfirst Int.compare) nos
          val nos = map #2 nos


          val nodetype = Vector.fromList nos

          val map_ = topregraph (nodes, edges)

          val pg = 
          { 
            (* orig_nodes = nodes,
            orig_edges = edges, *)
            name_to_nid_map = name_to_nid_map,
            nid_to_name_map = nid_to_name_map,
            map = map_,
            nelements = length nodes,

            nodetype = nodetype,

            hq = 
            case VectorUtil.findi (fn Types.Hq => true 
                                   | _ => false) nodetype of
                NONE => raise Graph "no hq?"
              | SOME (i, _) => i,

            robberstart = 
            case VectorUtil.findi (fn Types.RobberStart => true 
                                   |  _ => false) nodetype of
                NONE => raise Graph "no robberstart?"
              | SOME (i, _) => i,

            banks = 
            List.mapPartial (fn x => x)
            (ListUtil.mapi (fn (Types.Bank, i) => SOME i | _ => NONE)
             (Vector.foldr op:: nil nodetype))

            } : pregraph

          val paths = 
              Vector.tabulate (n,
                               fn me =>
                               let
                                 val paths_copfoot = dijkstra n me (Only Types.CopFoot) pg
                                 val paths_copcar = dijkstra n me (Only Types.CopCar) pg
                                 (* XXX PERF this one is probably redudant *)
                                 val paths_robber = dijkstra n me (Only Types.Robber) pg
                                 val paths_anyfoot = dijkstra n me (Any Types.CopFoot) pg
                                 val paths_anycar = dijkstra n me (Any Types.CopCar) pg

                                 fun f (Only Types.CopFoot) = paths_copfoot
                                   | f (Only Types.CopCar) = paths_copcar
                                   | f (Only Types.Robber) = paths_robber
                                   | f (Any Types.CopFoot) = paths_anyfoot
                                   | f (Any Types.CopCar) = paths_anycar
                                   | f (Any Types.Robber) = raise Graph "didn't expect Any Robber"
                               in
                                   f : transport -> SI.int Vector.vector
                               end)
      in
          { pregraph = pg,
            paths = paths} : graph
      end

  fun nodetype ({pregraph = {nodetype, ...}, ...} : graph) i =
      Vector.sub(nodetype, i)

  fun banks ({pregraph = {banks, ...}, ...} : graph) = banks
  fun hq ({pregraph = {hq, ...}, ...} : graph) = hq
  fun robberstart ({pregraph = {robberstart, ...}, ...} : graph) = robberstart


  fun orig_edges (graph : graph) =
      NM.foldli (fn (src, {out_edges,...}, lst) =>
                    foldl (fn ((dest, edge_tag), acc) =>
                              (src, dest, edge_tag) :: acc)
                          []
                          out_edges)
                []
                (#map (#pregraph graph))

  fun orig_nodes (graph : graph) =
      Vector.foldri (fn (nid, nodetag, lst) => (nid, nodetag, 0, 0) :: lst)
                    []
                    (#nodetype (#pregraph graph))

  fun length (g : graph) = #nelements (#pregraph g)

  fun adjacent ({pregraph = pg as {map, ...}, ...} : graph) dir trans from = 
      adjacent_pre pg dir trans from

  fun reachable (graph : graph) dir trans froms 0 = froms
    | reachable (graph : graph) dir trans froms n = 
      let fun adj nds = List.map (fn (nd, _) => nd) 
                                 (adjacent graph dir trans nds)
          val result = List.foldl (fn (from, reach) => union (reach, adj from))
                                  froms froms
      in
        reachable graph dir trans result (n-1)
      end

  fun distance ({pregraph, paths, ...} : graph) transport src dest =
    if src = dest 
    then 0
    else SI.toInt (Vector.sub((Vector.sub(paths, src) transport), dest))

  fun find_shortest_path graph transport src dest =
      if src = dest then nil else 
      let val steps = adjacent graph Forward transport src
          fun score (nd, pt) = 
              let val transport = case transport of Only pt => Only pt
                                                  | Any _ => Any pt
              in
                (distance graph transport nd dest, (nd, pt))
              end

          val V = List.map score steps
          val (_, (nd, pt)) = List.foldl (fn ((v, (nd, pt)), 
                                              (bv, (bnd, bpt))) =>
                                             if (v >= 0) andalso (v < bv) 
                                             then (v, (nd, pt))
                                             else (bv, (bnd, bpt)))
                                         (999, (0, Types.Robber)) V
          val transport = case transport of Only pt => Only pt
                                          | Any _ => Any pt
      in 
        (nd, pt) :: find_shortest_path graph transport nd dest
      end
      
  fun shortest_path_distance (graph as {paths, ...} : graph) 
                               transport src dest =
       if src = dest 
       then SOME (0, nil)
       else 
         let val d = SI.toInt (Vector.sub((Vector.sub(paths, src) transport), dest))
         in
           if d < 0 
           then NONE 
           else SOME (d, 
                      find_shortest_path graph 
                      transport src dest)
         end

  fun shortest_path g t s d = 
         Option.map (fn (_, x) => x) (shortest_path_distance g t s d)

end
