structure Parsing =
struct

open Message
exception Parse of string

fun send (s) = (print s; print "\n")
fun recv () = let val s = valOf (TextIO.inputLine TextIO.stdIn)
                  val _ = Spew.echo s
              in s end

fun unparse_ptype CopFoot = "cop-foot"
  | unparse_ptype CopCar = "cop-car"
  | unparse_ptype Robber = "robber"

fun parse_ptype "cop-foot" = CopFoot
  | parse_ptype "cop-car" = CopCar
  | parse_ptype "robber" = Robber
  | parse_ptype _ = raise Parse "expected ptype"

fun unparse_node_tag Hq = "hq"
  | unparse_node_tag Bank = "bank"
  | unparse_node_tag RobberStart = "robber-start"
  | unparse_node_tag Ordinary = "ordinary"

fun parse_node_tag "hq" = Hq
  | parse_node_tag "bank" = Bank
  | parse_node_tag "robber-start" = RobberStart
  | parse_node_tag "ordinary" = Ordinary
  | parse_node_tag _ = raise Parse "expected node_tag"

fun parse_edge_tag "car" = Car
  | parse_edge_tag "foot" = Foot
  | parse_edge_tag _ = raise Parse "expected edge_tag"

fun unparse_register_msg (name, ptype) =
    "reg: " ^ name ^ " " ^ unparse_ptype ptype

fun unparse_single_move_msg (loc, ptype) = 
    "mov: " ^ loc ^ " " ^ (unparse_ptype ptype)

fun unparse_list name (unparse_elt : 'a -> string) list =
    name ^ "\\" ^"\n"^
    (String.concat (map (fn x => name ^ ": " ^ (unparse_elt x) ^"\n") list)) ^
    name ^ "/"

fun unparse_move (loc, ptype, bot) = loc ^" "^ (unparse_ptype ptype) ^" "^ bot

fun unparse_move_msg moves = unparse_list "mov" unparse_move moves

fun unparse_clean_cop_move_msg {moves, offer, accuse} =
    let fun unparse_offer Turncoat = "turncoat:"
          | unparse_offer StraightArrow = "straight-arrow:"
    in
        "cmov\\\n" ^
        (unparse_offer offer) ^"\n"^
        (unparse_move_msg moves) ^"\n"^
        (unparse_list "acc" (fn bot => bot) accuse) ^"\n"^
        "cmov/"
    end

fun unparse_dirty_cop_move_msg moves = unparse_move_msg moves

fun unparse_robber_move_msg {move, result} =
    let fun unparse_bribe_result (Push (bot, loc)) = "psh: " ^ bot ^" "^ loc
          | unparse_bribe_result (BribeChoose cops) =
               unparse_list "chc" (fn cop => cop) cops
          | unparse_bribe_result BribeNone = "nobribe:"
    in
        "rmov\\\n" ^
        "mov: " ^ (unparse_move move) ^"\n"^
        (unparse_bribe_result result) ^"\n"^
        "rmov/"
    end

fun unparse_bribe_msg Bribe = "bribe:"
  | unparse_bribe_msg NoBribe = "nobribe:"

fun isWS #" " = true
  | isWS #"\t" = true
  | isWS #"\n" = true
  | isWS #"\r" = true
  | isWS _ = false

fun isNL "\n" = true
  | isNL "\r\n" = true
  | isNL _ = false

fun tokens s =
    if s = "" then
        []
    else
        let val (tok, s) = StringUtil.token isWS s
        in
            tok :: tokens s
        end

fun recv_line s =
    let val line = recv ()
        val toks = tokens line
    in
        if toks = [s] then
            ()
        else
            raise Parse ("expected \"" ^ s ^ "\"")
    end

fun recv_list_rest name (recv_elt : string list -> 'a) =
    let val finish = name ^ "/"
        val delim = name ^ ":"
        fun loop () =
            case tokens(recv()) of
                [tok] =>
                  if tok = finish then
                      []
                  else
                      raise Parse ("Expected " ^ finish)
              | d :: rest =>
                  if d = delim then
                      (recv_elt rest) :: loop()
                  else
                      raise Parse ("Expected " ^ delim)
              | _ => raise Parse ("Expected " ^ delim ^ " or " ^ finish)
    in
        loop() 
    end

fun recv_list name (recv_elt : string list -> 'a) =
    let val start = name ^ "\\"
        val () = recv_line start
    in recv_list_rest name recv_elt
    end

fun recv_nodes () =
    let fun get_node (lst : string list) =
            case lst of 
                [loc_s, node_tag_s, coord1_s, coord2_s] =>
                (* XXX maybe verify that "name"s are really "name"s? *)
                    (loc_s, parse_node_tag node_tag_s,
                     valOf (Int.fromString coord1_s),
                     valOf (Int.fromString coord2_s))
              | _ => raise Parse ("recv_nodes: incorrect # of args")
    in
        recv_list "nod" get_node
    end

fun recv_edges () =
    let fun get_edge toks = 
            case toks of
                  [loc1_s, loc2_s, edge_tag_s] =>
                  (loc1_s, loc2_s, parse_edge_tag edge_tag_s)
                | _ => raise Parse "recv_edges: wrong # of args"
    in
        recv_list "edg" get_edge
    end

fun unparse_one_string start line =
    case tokens line of
        [st, string] =>
            if start = st then
                string
            else
                raise Parse ("expected \"" ^ start ^ "\"")
      | _ => raise Parse ("expected \"" ^ start ^ "\"")

fun unparse_one_int start line =
    valOf (Int.fromString (unparse_one_string start line))

fun recv_world_skeleton () : Message.world_skeleton_msg =
    let val () = recv_line "wsk\\"
        val name = unparse_one_string "name:" (recv ())
        val robber = unparse_one_string "robber:" (recv ())
        val cop1 = unparse_one_string "cop:" (recv ())
        val cop2 = unparse_one_string "cop:" (recv ())
        val cop3 = unparse_one_string "cop:" (recv ())
        val cop4 = unparse_one_string "cop:" (recv ())
        val cop5 = unparse_one_string "cop:" (recv ())
        val nodes = recv_nodes ()
        val edges = recv_edges ()
        val () = recv_line "wsk/"
    in
        { name = name,
          robber = robber,
          cop = (cop1, cop2, cop3, cop4, cop5),
          graph = Graph.tograph (nodes, edges) }
    end

fun recv_dirty_cops () =
    let fun recv_dirty_cop toks = 
            case toks of
                  [brain] => brain
                | _ => raise Parse "recv_dirty_cop: wrong # of args"
    in
        recv_list "dc" recv_dirty_cop
    end

fun recv_controlled_cops () =
    let fun recv_controlled_cop toks = 
            case toks of
                  [brain, bot] => (brain, bot)
                | _ => raise Parse "recv_controlled_cop: wrong # of args"
    in
        recv_list "sc" recv_controlled_cop
    end

fun recv_false_accusations () =
    let fun recv_false_accusation toks = 
            case toks of
                  [accuser, accused, world] => (accuser, accused, 
                                                valOf (Int.fromString world))
                | _ => raise Parse "recv_false_accusation: wrong # of args"
    in
        recv_list "fac" recv_false_accusation
    end

fun recv_bank_vals () =
    let fun recv_bank_val toks = 
            case toks of
                  [loc, bv_s] =>
                  (* Check bv_s between 0 and 1000 *)
                  (loc, valOf (Int.fromString bv_s))
                | _ => raise Parse "recv_bank_val: wrong # of args"
    in
        recv_list "bv" recv_bank_val
    end

fun recv_evidence () =
    let fun recv_ev toks = 
            case toks of
                [loc, world_s] =>
                (* Check world between 0 and 200 *)
                    (loc, valOf (Int.fromString world_s))
              | _ => raise Parse "recv_evidence: wrong # of args"
    in
        recv_list "ev" recv_ev
    end

fun recv_player_locs () = 
    let fun recv_player_loc toks =
            case toks of
                [bot, loc, ptype_s] => (bot, loc, parse_ptype ptype_s)
              | _ => raise Parse "recv_player_locs: wrong # of args"
    in
        recv_list "pl" recv_player_loc
    end

fun recv_offered_cops () =
    case tokens (recv ()) of
        ["nottelling:"] => NotTelling
      | ["ofc\\"] =>
            let fun recv_offered_cop toks =
                case toks of
                    [bot] => bot
                  | _ => raise Parse "recv_offered_cops: wrong # of args"
                val cops = recv_list_rest "ofc" recv_offered_cop
            in
                Offered cops
            end
      | _ => raise Parse "expected \"nottelling\" or \"ofc\\\""

fun recv_world_or_game_over () =
    case tokens (recv ()) of 
        ["game-over"] => (TextIO.flushOut TextIO.stdErr; NONE)
      | ["wor\\"] =>
            let val world = unparse_one_int "wor:" (recv ())
                val rbd = unparse_one_int "rbd:" (recv ())
                val dc = recv_dirty_cops ()
                val sc = recv_controlled_cops ()
                val fac = recv_false_accusations ()
                val bank_vals = recv_bank_vals ()
                val evidence = recv_evidence ()
                val smell = case unparse_one_int "smell:" (recv ()) of
                                0 => NONE
                              | n => SOME n
                val player_locs = recv_player_locs ()
                val () = recv_line "wor/"
            in
                SOME { time = world,
                       loot = rbd,
                       dirty_cops = dc,
                       controlling = sc,
                       accused = fac,
                       bank_vals = bank_vals,
                       evidence = evidence,
                       smell = smell,
                       player_locs = player_locs }
            end
      | ["accused:"] => ((*TextIO.flushOut TextIO.stdErr; *) NONE)
      | _ => raise Parse "expected \"game-over\" or \"wor\\\" or \"accused\""

end
