
(* The robber predicts the future by predicting "heat", which is a
   probability distribution of cops.

   Since the robber knows where cops are when he starts doing search,
   the heat is totally localized into the five cop locations. After
   our first move, however, we can't predict precisely where they'll
   be (and since their movements will generally be based on probabilistic
   information, and there are so many possible moves, it seems implausible
   to explore the discrete state space directly).

   Heat transfer simultaneously does transfer for everyone. Cops in
   this setting always make the same "move" (which is to split up
   their probability into surrounding spots), based on what they know
   about where the robber is.

   But how is "knowledge" represented in the probability distribution?
   For each (occupied) node (cross vehicle), we compute a slime belief
   network that represents what a cop in that position believes.

   The belief network is a prediction of "slime": a probability
   distribution for the robber. The belief network is different for each
   node because it depends on smells and evidence collected 

   This doesn't use information perfectly. For instance, two cops with
   very different slimenets that arrive in the same node will necessarily
   now need to share a slimenet, which will be some sort of average of
   the two.

   To summarize:

   'heat' describes belief about the location of cops (and what they
   believe) and what vehicles they're riding. This is represented
   as a distribution over states, along with a slime belief network
   for each one state*vehicle.

   'slime' is a distribution over states that indicates a cop's belief
   that a robber is in that state.

*)

(* omg am I programmingInJava? *)
structure RobberHeatTransform =
struct

    datatype coptrans = ROBOCAR | ROBOFOOT

    (* we get SLIMEFACTOR * n_nodes worth of slime to
       distribute. *)
    val SLIMEFACTOR = 4
    (* number of robber hypotheses to consider *)
    val NHYPS = 5

    open Transform
    
    type slime = 
        (* length Graph.length *)
        int Vector.vector

    (* representation of the cops, as a heat distribution *)
    type cops_asheat =
        (* length of each vector is Graph.length *)
              (* weight>0, slime *)
        { foot : (int * slime) option Vector.vector,
          car : (int * slime) option Vector.vector }
        
    (* do a cop turn for all cops. *)
    fun transform (GS{ skel, time, loot, bank_vals, 
                       evidence, cops = heat : cops_asheat, 
                       robber : exactrobber, transfers,
                       temp })
        : (cops_asheat, exactrobber) Transform.gamestate
        =
     if time mod 2 <> 1
     then raise Transform "cops transform odd worlds."
     else 
     let 

       val zeroslime = Vector.tabulate(Graph.length skel, fn _ => 0)

       (* step 1.
          update slime.

          we might see evidence (then we restrict our
          slime to a circle around the evidence), or
          smell (same)
          or we might see a robbery (then we make all
          the slime in one location).
          *)

       (* restrict slime to a radius (in terms of graph, the
          reachable nodes with 'radius' bot moves) of a center.
          any other slime goes to zero, but the area is normalized
          to make it sum to the same old value 
          (always SLIMEFACTOR*Graph.length skel).

          we could use reachable, and that would probably be faster for
          small radiuses, but if the radius is like 8, then we would
          be computing all paths that are 8 long, which is wayyyy loads. *)
       fun restrictslime center radius oldslime =
           let
               (* step a. find the total of anything within the radius *)
               val total_inrange = ref 0
               val _ = Util.for 0 (Graph.length skel - 1)
                   (fn x =>
                    if Graph.distance skel (Graph.Only Types.Robber) center x <= radius
                    then 
                        total_inrange :=
                          !total_inrange + Vector.sub(oldslime, x)
                    else ())
               val total_inrange = !total_inrange

               (* normalization factor is 1/percentage_already_in_radius.
                  (so if half of the slime is in the radius, we'll double
                  its value) *)
               val normalization_factor =
                   real (SLIMEFACTOR * Graph.length skel) /
                   (* fudge: avoid dbz *)
                   (real (total_inrange) + 0.01)
                   
               (* step b. make new vector where slime is 0 if outside
                  radius, or old_value * normalization_factor if inside *)
           in
               Vector.tabulate
               (Graph.length skel,
                (fn x =>
                 if Graph.distance skel (Graph.Only Types.Robber) center x <= radius
                 then 
                     (Real.trunc
                      ((0.01 + real (Vector.sub (oldslime, x))) * normalization_factor))
                     handle _ => 9999
                 else 0))
           end

       (* XXX also, any slime at the current location+smell radius should be
          redistributed elsewhere, since if we caught the robber, we
          wouldn't be here. *)
       fun maybe_smellvidence heat = 
           let
               fun upslime _ (_, NONE) = NONE
                 | upslime pt (n, SOME (ncops, oldslime)) =
                   (* smell range? *)
                   let 
                       val range = 
                           case pt of
                               Types.CopFoot => 2
                             | _ => 1

                   in
                       (* XXX isn't extra-smart when receiving evidence and
                          smell in the same turn, oh well *)
                       SOME 
                       (ncops,
                        if Graph.distance skel (Graph.Only pt) n robber <= range
                        then (* smelled! *)
                            restrictslime n range oldslime 
                        else 
                            (* evidence? *)
                            (case ListUtil.Alist.find op= evidence n of
                                 SOME when => 
                                     restrictslime n (time - when) oldslime
                               | NONE => 
                                   (* firstround? *)
                                     if time = 1
                                     then 
                                         restrictslime 
                                           (Graph.robberstart skel) 1 oldslime
                                     else oldslime))
                   end

           in
               { foot = Vector.mapi (upslime Types.CopFoot) (#foot heat),
                 car  = Vector.mapi (upslime Types.CopCar) (#car heat) }
           end
           

       (* note: we can't 'pick up' evidence since we don't
          have true cops here, I guess we should just leave it
          or else support probabilistic evidence too. *)

       val heat =
           if Graph.nodetype skel robber = Types.Bank
           then (* heat converges *)
               let
                   val perfect =
                       Vector.tabulate
                       (Graph.length skel,
                        fn i => if i = robber 
                                then SLIMEFACTOR * Graph.length skel
                                else 0)

                   fun upslime NONE = NONE
                     | upslime (SOME (ncops, _)) = 
                       SOME (ncops, perfect)
               in
                   { foot = Vector.map upslime (#foot heat),
                     car = Vector.map upslime (#car heat) }
               end
           else maybe_smellvidence heat

       (* XXX step 2.
          allow some cross-over of slime, from voting
          round (assume voting is somewhat successful) *)

       (* step 3. transfer heat and slime *)

       (* get the step that gets us to the dest most quickly *)
       fun get_step nowtransport src dest 
           : Graph.nid * coptrans =
           let
               val pt = 
                   (case nowtransport of
                        ROBOFOOT => Types.CopFoot
                      | ROBOCAR => Types.CopCar)
               val reachable =
                   (src, pt) ::
                   Graph.adjacent
                     skel Graph.Forward
                     (Graph.Any pt)
                     src

               val n = length reachable


               val dists =
                   map (fn (src, pt) =>
                        (src, pt, Graph.distance skel (Graph.Any pt) src dest))
                       reachable

               val (choice, how, _) = 
                   ListUtil.min (fn ((_, _, d1),
                                     (_, _, d2)) =>
                                         Int.compare (d1, d2))
                                dists
           in
               (choice, case how of Types.CopFoot => ROBOFOOT
                               | _ => ROBOCAR)
           end handle e => raise e

       (* we assume cops head towards slimy areas. *)
       val heat = 
           let
               val newheat_car =
                   Array.array (Graph.length skel,
                                NONE)

               val newheat_foot =
                   Array.array (Graph.length skel,
                                NONE)

               (* spread from each i *)
               fun spread vec transport src =
                   case Vector.sub(vec, src) of
                       NONE => () (* no cops, so do nothing *)
                     | SOME (cops, slime) => 
                       let 

                           (* go towards the most slimy
                              areas. pick NHYPS hypotheses
                              about where the robber is,
                              then move towards them based
                              on their slimy ranks. *)

                           (* in order to rank hypotheses,
                              we could discount them by their
                              proximity to the current location *)
                           fun look nl i =
                               if i = Vector.length slime
                               then nl
                               else
                                   let
                                       val sl =
                                           Vector.sub(slime, i)

                                       val nl =
                                           ListUtil.Sorted.insertbest
                                           NHYPS
                                           (ListUtil.bysecond
                                            (ListUtil.Sorted.reverse
                                             Int.compare))
                                           nl
                                           (* never consider zero
                                              slime, since it gives
                                              us dbz errors *)
                                           (i, sl + 1)

                                   in
                                       look nl (i + 1)
                                   end handle e => raise e

                           val hyps = look nil 0


                           (* now split the heat into cops following
                              those hypotheses *)

                           val total_slime =
                               foldl (fn ((_, s), b) => s + b)
                                  0 hyps

                           (* move cops in most productive vector,
                              in proportion to the amount of slime
                              at the target *)

                           (* add the number of cops assigned for
                              each hyp.. XX roundoff? *)
                           val hyps =
                               map (fn (target, sl) =>
                                    { target = target,
                                      step = get_step transport src target,
                                      sl = sl,
                                      assigned = 
                                      Real.trunc 
                                      (real cops * 
                                       (real sl / real total_slime)) })
                               hyps


                           fun doh { assigned = 0, step, sl, target } = ()
                             | doh { step = (dst, trans), 
                                     assigned, ... } =
                               let 
                                   val a = 
                                       case trans of 
                                           ROBOCAR => newheat_car
                                         | ROBOFOOT => newheat_foot
                                                   
                                   (* as we do the transfer,
                                      we accumulate with other sources *)
                                   val (curcops, scaledslime) =
                                       case Array.sub(a, dst) of
                                           NONE => (0, zeroslime)
                                         | SOME { newcops,
                                                  scaledslime } =>
                                               (newcops, scaledslime)
                               in
                                   Array.update
                                   (a, dst,
                                    SOME { newcops = curcops + assigned,
                                          
                                           scaledslime = 
                                          Vector.tabulate
                                          (Vector.length slime,
                                           (fn i =>
                                            Vector.sub(slime, i) * assigned +
                                            Vector.sub(scaledslime, i))) })
                               end handle e => raise e
                       in
                           (* now fill it in. the slime is scaled by the
                              total number of cops currently there, so
                              we need to normalize it at the end. 

                              invt: if SOME then newcops>0
                              *)
                           app doh hyps
                       end handle e => raise e

               fun normalize a =
                   Vector.tabulate(Graph.length skel,
                                   (fn i =>
                                    case Array.sub(a, i) of
                                        NONE => NONE
                                      | SOME { newcops, scaledslime } => 
                                            SOME
                                            (newcops,
                                             Vector.map (fn x =>
                                                         x div newcops)
                                             scaledslime)))
                   handle e => raise e

               (* for any slime vector, move most of the slime
                  into adjacent spots *)
               fun advance_slime v =
                   let
                       val na = Array.array(Vector.length v, 0)
                   in
                       Util.for 0 (Vector.length v - 1)
                       (fn i =>
                        let
                            val sl = Vector.sub(v, i)
                            val nbors = 
                                i ::
                                map #1
                                (Graph.adjacent skel Graph.Forward
                                 (Graph.Only Types.Robber)
                                 i)

                            val num = length nbors
                        in
                            app (fn j =>
                                 Array.update(na,
                                              j,
                                              Array.sub(na, j) +
                                              sl div num))
                        end);
                       VectorUtil.atov na
                   end handle e => raise e

               fun advance v =
                   Vector.map
                   (fn NONE => NONE
                     | SOME (cops, slime) =>
                          SOME (cops, advance_slime slime)) v

           in
               Util.for 0 (Vector.length (#foot heat) - 1)
                 (fn i => spread (#foot heat) ROBOFOOT i);

               Util.for 0 (Vector.length (#car heat) - 1)
                 (fn i => spread (#car heat) ROBOCAR i);

               { foot = advance (normalize newheat_foot),
                 car = advance (normalize newheat_car) }
           end

       val time = time + 1

       fun heatat v i =
           case Vector.sub(v, i) of
               NONE => 0
             | SOME (h, _) => h
     in
         Transform.GS
         { skel = skel,
           time = time,
           loot = loot,
           temp = temp +
                  heatat (#car heat) robber +
                  heatat (#foot heat) robber,
           bank_vals = bank_vals,
           evidence = evidence,
           cops = heat,
           robber = robber,
           transfers = transfers }
     end

    val transform = fn x => (transform x handle e => raise e)

end
