structure Super :> SUPER =
struct

  type pid = Types.pid
  type bid = Types.bid
  type node = Types.node
  type edge = Types.edge
  type time = Types.time
  type nid = Types.nid
  type ptype = Types.ptype

  type register_msg = pid * ptype

  type world_skeleton_msg  = {name : pid, 
			      robber: pid,
			      cop : pid * pid * pid * pid * pid, 
                              graph : Graph.graph }

  type world_msg = {time: time,
		    loot: int,
                    dirty_cops : bid list,
                    controlling : (bid * pid) list,
                    accused : (bid * bid * time) list,
		    bank_vals: (nid * int) list,
		    evidence: (nid * time) list,
		    smell: int option,
		    player_locs : (pid * nid * ptype) list}

  type move_msg = nid * ptype * pid

  datatype offer = datatype Message.offer

  type clean_cop_move_msg = {moves: move_msg list,
                             offer : offer,
                             accuse : pid list}

  type dirty_cop_move_msg = move_msg list

  type accused = unit

  datatype bribe_msg = datatype Message.bribe_msg

  datatype offered_cops_msg = datatype Message.offered_cops_msg

  datatype bribe_result = Push of pid * nid
                        | BribeChoose of bid list
                        (* Use if robber said nobribe in previous msg. *)
                        | BribeNone

  type robber_move_msg = {move : nid,
                          result : bribe_result}

  type inform_msg = (pid * nid * ptype * time * int) list

  type plan_msg = (pid * nid * ptype * time) list

  type vote_msg = bid list

  type vote_res_msg = bid option

  type from_inform_msg = (pid * inform_msg) list
  type from_plan_msg = (bid * plan_msg) list
	               
  type state = exn

  structure G = Graph

  type graph = G.graph

  type super = {name : pid, robber : pid, cops : pid * pid * pid * pid * pid,
                graph : graph,
                time : time, loot : int, evidence : (nid * time) list,
                smell : int option,
                player_locs : (pid * nid * ptype) list, 
                dirty_cops : bid list,
                controlling : (bid * bid) list,
                informs : from_inform_msg, plans : from_plan_msg}

  exception Super of string

  fun get_name ({name, ...} : super) = name
  fun get_robber ({robber, ...} : super) = robber
  fun get_cops ({cops, ...} : super) = cops
  fun get_cops_as_list ({cops=(a, b, c, d, e), ...} : super) = [a, b, c, d, e]
  fun get_nodes ({graph, ...} : super) = G.orig_nodes graph
  fun get_banks_as_list ({graph, ...} : super) 
     = List.filter (fn (_, ntag, _, _) => ntag = Message.Bank) (G.orig_nodes graph)
  fun get_edges ({graph, ...} : super) = G.orig_edges graph
  fun get_graph ({graph, ...} : super) = graph

  fun get_time ({time, ...} : super) = time
  fun get_loot ({loot, ...} : super) = loot  
  fun get_dirty_cops ({dirty_cops, ...} : super) = dirty_cops
  fun get_controlling ({controlling, ...} : super) = controlling
  fun get_smell ({smell, ...} : super) = smell  
  fun get_evidence ({evidence, ...} : super) = evidence

  fun get_player_locs ({player_locs, ...} : super) = player_locs
  fun get_cop_locs ({player_locs, ...} : super) =
    List.filter (fn (pid, nid, ptype) => ptype <> Types.Robber) player_locs
  fun get_player_loc ({player_locs, ...} : super, pid) 
      = (case List.find (fn (pid', _, _) => pid = pid') player_locs
          of SOME (_, nd, pt) => (nd, pt)
           | NONE => raise Super ("didn't find player " ^ pid))
  fun get_my_loc u = get_player_loc (u, get_name u)

  fun get_informs ({informs, ...} : super) = informs
  fun get_informs_by_informer ({informs, ...} : super, pid)
      = (case List.find (fn (pid', _) => pid = pid') informs
          of SOME (_, infms) => infms
           | NONE => raise Super ("didn't find informer " ^ pid))

  fun get_plans ({plans, ...} : super) = plans
  fun get_plans_by_planner ({plans, ...} : super, pid)
      = (case List.find (fn (pid', _) => pid = pid') plans
          of SOME (_, plns) => plns
           | NONE => raise Super ("didn't find planner " ^ pid))

  type dirtycop = {
    inform : super * world_msg * state -> inform_msg * state,
    plan : super * from_inform_msg * state -> plan_msg * state,
    vote : super * from_plan_msg * state -> vote_msg * state,

    move : super * vote_res_msg * state -> dirty_cop_move_msg * state,

    (* For forming dastardly plans! *)
    r_inform : super * world_msg * state -> inform_msg * state,
    r_plan : super * from_inform_msg * state -> plan_msg * state,
    r_vote : super * from_plan_msg * state -> vote_msg * state,
    r_results : super * vote_res_msg * state -> state
  }

  type cleancop = {
    register : unit -> register_msg * state,
    world_skeleton : super * world_skeleton_msg * state -> state,

    dirtify : super * state -> dirtycop * state,

    inform : super * world_msg * state -> inform_msg * state,
    plan : super * from_inform_msg * state -> plan_msg * state,
    vote : super * from_plan_msg * state -> vote_msg * state,

    move : super * vote_res_msg * state -> clean_cop_move_msg * state
  }

  type robber = {
    register : unit -> register_msg * state,
    world_skeleton : super * world_skeleton_msg * state -> state,

    inform : super * world_msg * state -> inform_msg * state,
    plan : super * from_inform_msg * state -> plan_msg * state,
    vote : super * from_plan_msg * state -> vote_msg * state,
    bribe : super * vote_res_msg * state -> bribe_msg * state,
    move : super * offered_cops_msg * state -> robber_move_msg * state
  }

  val BOTNAME = Params.param "-" (SOME ("-botname",
                                        "Force a name for this bot"))
                             "botname"
                             
  exception Impossible

  exception P of super * exn

  fun sregister f () =
      let val ((pid, pt), s) = f ()
          val u = {name="?", robber="?", cops=("?","?","?","?","?"),
                   graph=G.dummy, 
                   time=0, loot=0, smell=NONE,
                   evidence=[],
                   player_locs=nil, informs=nil, plans=nil,
                   dirty_cops=nil, controlling=nil}
          val pid = case !BOTNAME of "?" => pid | x => x
      in 
        ((pid, pt), P(u, s))
      end

  fun sworld_skeleton f (ws_msg : Message.world_skeleton_msg, 
                         P({time, loot, smell, evidence, player_locs, 
                            informs, plans, dirty_cops, controlling, ...}, s)) : state
    = let val name = #name ws_msg
          val robber = #robber ws_msg
          val cops = #cop ws_msg
          val graph = #graph ws_msg
          val u = {name=name, robber=robber, cops=cops,
                   graph=graph,
                   time=time, loot=loot, smell=smell, evidence=evidence,
                   player_locs=player_locs, informs=informs, plans=plans,
                   dirty_cops=dirty_cops, controlling=controlling}
          val s = f (u, ws_msg, s)
      in
        P(u, s)
      end
    | sworld_skeleton _ _ = raise Impossible

  fun sinform f (msg : Message.world_msg,
              P({name, robber, cops, graph,
                 plans, informs, dirty_cops, controlling, ...}, s))
      : Message.inform_msg * state
    = let val time = #time msg
          val loot = #loot msg
          val dirty_cops = #dirty_cops msg
          val controlling = #controlling msg
          val accused = #accused msg
          val smell = #smell msg
          val evidence = List.map (fn (name, v) => (G.name_to_nid graph name, v)) (#evidence msg)
          val player_locs = List.map (fn (pid, name, pt) => (pid, G.name_to_nid graph name, pt)) (#player_locs msg)

          val msg = {time=time, loot=loot, dirty_cops=dirty_cops, controlling=controlling, accused=accused, bank_vals=(List.map (fn (name, v) => (G.name_to_nid graph name, v)) (#bank_vals msg)),
           evidence=(*(List.map (fn (name, v) => (G.name_to_nid graph name, v)) (#evidence msg)),*)evidence,
           smell=smell,
           player_locs=player_locs}

          val u = {name=name, robber=robber, cops=cops,
                   graph=graph, evidence=evidence,
                   time=time, loot=loot, smell=smell,
                   player_locs=player_locs, informs=informs, plans=plans,
                   dirty_cops=dirty_cops, controlling=controlling}

          val (msg, s) = f (u, msg, s)
          val msg = List.map (fn (pd, nd, pt, t, c) => (pd, G.nid_to_name graph nd, pt, t, c)) msg
      in
        (msg, P(u, s))
      end
    | sinform _ _ = raise Impossible

  fun splan f (msg : Message.from_inform_msg,
               P({name, robber, cops, graph, 
                  time, loot, smell, player_locs, 
                  plans, evidence, dirty_cops, controlling, ...}, s))
      : Message.plan_msg * state
    = let 
          val msg = List.map (fn (pd, imsg) => (pd, List.map (fn (pd, name, pt, t, c) => (pd, G.name_to_nid graph name, pt, t, c)) imsg)) msg

      val informs = msg
          val u = {name=name, robber=robber, cops=cops,
                   graph=graph,
                   time=time, loot=loot, smell=smell, evidence=evidence,
                   player_locs=player_locs, informs=informs, plans=plans,
                   dirty_cops=dirty_cops, controlling=controlling}

          val (msg, s) = f (u, msg, s)

          val msg = List.map (fn (pd, nd, pt, t) => (pd, G.nid_to_name graph nd, pt, t)) msg

      in
        (msg, P(u, s))
      end
    | splan _ _ = raise Impossible

  fun svote f (msg : Message.from_plan_msg,
              P({name, robber, cops, graph,
                 time, loot, smell, evidence, player_locs, 
                 informs, dirty_cops, controlling, ...}, s))
      : Message.vote_msg * state
    = let
          val msg = List.map (fn (pd, pmsg) => 
                                 (pd, List.map (fn (pd, name, pt, t) => (pd, G.name_to_nid graph name, pt, t)) pmsg)) msg

      val plans = msg
          val u = {name=name, robber=robber, cops=cops,
                   graph=graph,
                   time=time, loot=loot, smell=smell, evidence=evidence,
                   player_locs=player_locs, informs=informs, plans=plans,
                   dirty_cops=dirty_cops, controlling=controlling}

          val (msg, s) = f (u, msg, s)
      in
        (msg, P(u, s))
      end
    | svote _ _ = raise Impossible

  fun smove_cleancop f (msg : Message.vote_res_msg,
              P({name, robber, cops, graph,
                 time, loot, smell, evidence, player_locs, 
                 plans, informs, dirty_cops, controlling, ...}, s))
      : Message.clean_cop_move_msg * state
    = let val u = {name=name, robber=robber, cops=cops,
                   graph=graph,
                   time=time, loot=loot, smell=smell, evidence=evidence,
                   player_locs=player_locs, informs=informs, plans=plans,
                   dirty_cops=dirty_cops, controlling=controlling}
          val ({moves, offer, accuse}, s) = f (u, msg, s)
      in
        ({moves = List.map (fn (nd, pt, pd) => 
                               (G.nid_to_name graph nd, pt, pd)) moves,
          offer = offer,
          accuse = accuse}, P(u, s))
      end
    | smove_cleancop _ _ = raise Impossible

  fun smove_dirtycop f (vr_msg : Message.vote_res_msg, 
                        P(u as {graph, ...} : super, s))
      : Message.dirty_cop_move_msg * state
    = let val (moves, s) = f (u, vr_msg, s)
      in
        (List.map (fn (nd, pt, pd) => (G.nid_to_name graph nd, pt, pd)) moves,
         P (u, s))
      end
    | smove_dirtycop _ _ = raise Impossible

  fun sresults f (vr_msg : Message.vote_res_msg, P(u : super, s)) : state
    = let val s = f (u, vr_msg, s)
      in
        P (u, s)
      end
    | sresults _ _ = raise Impossible

  fun sdirtify f (P (u, s)) =
      let val ({inform, plan, vote, move, r_inform, r_plan, r_vote, r_results},
               s)
              = f (u, s)
      in
        ({inform = sinform inform,
          plan = splan plan,
          vote = svote vote,
          move = smove_dirtycop move,
          r_inform = sinform r_inform,
          r_plan = splan r_plan,
          r_vote = svote r_vote,
          r_results = sresults r_results}, P (u, s))
      end
    | sdirtify _ _ = raise Impossible

  fun sbribe (f : super * vote_res_msg * state -> bribe_msg * state)
            (vr_msg : Message.vote_res_msg, P (u as {graph, ...} : super, s))
     : Message.bribe_msg * state
     = let val (b_msg, s) : bribe_msg * state = f (u, vr_msg, s)
       in
         (b_msg, P (u, s))
       end
    | sbribe _ _ = raise Impossible

  fun smove_robber f (msg : Message.offered_cops_msg,
              P(u as {name, robber, cops, graph,
                 plans, informs, ...} : super, s))
      : Message.robber_move_msg * state
    = let fun map_bribe_result g (Push (pd, nd)) 
            = (Message.Push (pd, G.nid_to_name g nd))
            | map_bribe_result g (BribeChoose bds) = Message.BribeChoose bds
            | map_bribe_result g (BribeNone) = Message.BribeNone
          val ({move = nd, result}, s) = f (u, msg, s)
      in
        ({move = (G.nid_to_name graph nd, Types.Robber, name),
          result = map_bribe_result graph result}, P(u, s))
      end
    | smove_robber _ _ = raise Impossible

  fun super_cop ({register, world_skeleton, dirtify, inform, plan, vote, move}
                 : cleancop) 
      : Message.cleancop = {register = sregister register,
                            world_skeleton = sworld_skeleton world_skeleton,
                            dirtify = sdirtify dirtify,
                            inform = sinform inform,
                            plan = splan plan,
                            vote = svote vote,
                            move = smove_cleancop move}

  fun super_robber ({register, world_skeleton, inform, plan, vote, bribe, move}
                    : robber) 
      : Message.robber = {register = sregister register,
                          world_skeleton = sworld_skeleton world_skeleton,
                          inform = sinform inform,
                          plan = splan plan,
                          vote = svote vote,
                          bribe = sbribe bribe,
                          move = smove_robber move}

end
